"""
"""

from PyQt4.QtGui  import *
from PyQt4.QtCore import *

import Constants
import sys
import re

class Bridge (QObject):
	"""
	"""

	def __init__ (self, app, win):
		"""
		Overview
		========
		The bridge controls many "global" functionality, and in some cases may want to alter
		something in the main window or application.

		Also, and probably most importantly, the bridge serves as a universal interconnection
		between any class you want.  It makes it incredibly easy for the open file menu to send a
		command to the ipython shell which is going to perform a command and then update the
		region and variable viewers.

		@param app: This gives the bridge a reference to the underlying application, allowing
		            it to change its fundamental properties if necessary.
		@type app: QApplication

		@param win: This gives the bridge a reference to the underlying main window, allowing
		            it to change its fundamental properties if necessary.
		@type win: QMainWindow
		"""

		QObject.__init__ (self)
		self._app = app
		self._mainWindow = win

		self._connected = False
		self._parallel = False
		self._drugItem = None

		self._host = None
		self._port = None
		self._user = None
		self._nodes = []
		self._numNodes = 0

		# Imported libraries (get set in IPythonView.Init ())
		self._importPylab = False
		self._importNumpy = False
		self._importScipy = False

		try:
			import visit
			self._importVisIt = True
		except (ImportError):
			self._importVisIt = False


		# Defaults, before general.conf gets loaded
		self._inPlace = False
		self._autoGuess = True
		self._limitOut = True
		self._dataPath = None

		self._pythonVersion = sys.version_info

		self._visitModuleStarted = False

		# attempt to refresh ever 3 seconds
		self._lastUpdated = 0
		self.startTimer (3000)

		self.loadSettings ()

	def loadSettings (self):
		"""
		Overview
		========
		This method will load the contents of the configuration file: ~/.basin/client/general.conf

		These correspond to the general preferences found within the BASIN_preferences.Preferences class.

		This method is called upon the instantiation of the bridge as well as whenever the general.conf
		is overwritten by the preferences menu.
		"""

		file = QFile (str (QDir.home ().path ()) + "/.basin/client/general.conf")
		if file.exists ():
			file.open (QIODevice.ReadOnly | QIODevice.Text)
			stream = QTextStream (file)
			reInPlace = re.compile ("\s*INPLACE\s*:\s*.*\s*")
			reAutoGuess = re.compile ("\s*AUTOGUESS\s*:\s*.*\s*")
			reLimitOut = re.compile ("\s*LIMITOUT\s*:\s*.*\s*")
			reDataPath = re.compile ("\s*DATAPATH\s*:\s*.*\s*")
			while not stream.atEnd ():
				line = str (stream.readLine ())
				val = line[line.find (':') + 1:]
				if   val == "False":
					val = False
				elif val == "True":
					val = True
				elif val == "None":
					val = None

				if   reInPlace.match (line):
					self._inPlace = val
				elif reAutoGuess.match (line):
					self._autoGuess = val
				elif reLimitOut.match (line):
					self._limitOut= val
				elif reDataPath.match (line):
					self._dataPath = val
			file.close ()

	def setMenuBar (self, menuBar):
		"""
		Overview
		========
		This gives the bridge a reference to the menubar. This isn't done during initialization
		since the bridge is created before the menu bar exists.

		@param menuBar: This is the reference to the menu bar that the Bridge requires.
		@type menuBar: MenuBar
		"""

		self._menuBar = menuBar

	def setToolBar (self, toolBar):
		"""
		Overview
		========
		This gives the bridge a reference to the menubar. This isn't done during initialization
		since the bridge is created before the tool bar exists.

		@param toolBar: This is the reference to the tool bar that the Bridge requires.
		@type toolBar: ToolBar
		"""

		self._toolBar = toolBar

	def setActions (self, actions):
		"""
		Overview
		========
		This gives the bridge a reference to the actions. This isn't done during initialization
		since the bridge is created before the actions exists.

		@param actions:
		@type actions:
		"""

		self._actions = actions

	def setRegionViewer (self, regionViewer):
		"""
		Overview
		========
		This gives the bridge a reference to the region viewer. This isn't done during initialization
		since the bridge is created before the region viewer exists.

		@param regionViewer:
		@type regionViewer:
		"""

		self._regionViewer = regionViewer
		QObject.connect (self, SIGNAL ("updateBasin (PyQt_PyObject)"), self._regionViewer.Update)

	def setVariableViewer (self, varViewer):
		"""
		Overview
		========
		This gives the bridge a reference to the variable viewer. This isn't done during initialization
		since the bridge is created before the variable viewer exists.
		"""

		self._varViewer = varViewer
		QObject.connect (self, SIGNAL ("updateWorld (PyQt_PyObject)"), self._varViewer.Update)

	def setDock (self, dock):
		"""
		Overview
		========
		This gives the bridge a reference to the dock. This isn't done during initialization since
		the bridge is created before the dock exists.

		@param dock: 
		@type dock: Dock
		"""

		self._dock = dock

	def setStatusBar (self, statusBar):
		"""
		Overview
		========
		This gives the bridge a reference to the status bar. This isn't done during initialization
		since the bridge is created before the status bar exists.
		"""

		self._statusBar = statusBar

	def setIPyShell (self, ipyShell):
		"""
		Overview
		========
		This gives the bridge a reference to the IPython Shell. This isn't done during initialization
		since the bridge is created before the IPython Shell exists.
		"""

		self._ipyShell = ipyShell
		self.pythonStatus ()

	def CursorNorm (self):
		"""
		Overview
		========
		This method returns the cursor to its former state. This means that you shouldn't change
		a cursor until it has been returned to its original state, because returning it twice has no effect.
		"""

		self._app.restoreOverrideCursor ()

	def CursorBusy (self):
		"""
		Bridge.CursorBusy (self)

		This method changes the cursor to a busy cursor.  This
		should really only be called if the cursor is in a 
		nromal state to begin with.
		"""
		self._app.setOverrideCursor (QCursor (Qt.WaitCursor))

	def CursorOpenHand (self):
		"""
		Bridge.CursorBusy (self)

		This method changes the cursor to an open hand cursor.
		This should really only be called if the cursor is in a 
		nromal state to begin with.
		"""
		self._app.setOverrideCursor (QCursor (Qt.OpenHandCursor))

	def CursorClosedHand (self):
		"""
		Bridge.CursorBusy (self)

		This method changes the cursor to a closed hand cursor.
		This should really only be called if the cursor is in a 
		nromal state to begin with.
		"""
		self._app.setOverrideCursor (QCursor (Qt.ClosedHandCursor))
		

	def Update (self):
		"""
		Bridge.Update (self)

		This is the updating function that should be used to perform
		an update across the entie basin_client.  While the individual
		region viewer and free variable viewer have update methods,
		this should be the only point at which they are called.

		That being said, the roder in which they are called is of
		great importance.  The Region Viewer must be updated first,
		then the free variable viewer.  This is because the variable
		viewer has to check for aliases after it sets its cotnents up,
		and if the region viewer is not already updated, then it
		would be providing possibly false, and certainly old, results.

		As this operation can take some noticeable time, the cursor
		is changed to busy during this oepration.

		After the update is complete, the _lastUpdated member is also
		updated to reflect the ipython line number at which the client
		is updated to.  This gets used because every few seconds a
		timer checks if an update is needed, by comparing this _lastUpdated
		member to the current line.
		"""
		if (not self._connected) and (not self._parallel):
			return
		self.CursorBusy ()
		self.emit (SIGNAL ("updateBasin (PyQt_PyObject)"), self._ipyShell.getTopLevelRegions ())
		self.emit (SIGNAL ("updateWorld (PyQt_PyObject)"), self._ipyShell.getFreeVariables ())
		discrete, discreteWindows = self._ipyShell.getVisItExposeDiscrete ()
		continuous, continuousWindows = self._ipyShell.getVisItExposeContinuous ()
		self.emit (SIGNAL ("updateVisIt (PyQt_PyObject, PyQt_PyObject, PyQt_PyObject, PyQt_PyObject)"), discrete, discreteWindows, continuous, continuousWindows)
		self.CursorNorm ()
		self._ipyShell._commandLine._update == False
		result = self._ipyShell.SilentCommand ("print \"\\n\" + BASIN_UPDATE").split ("\n")
		lineNumber =  result[1][result[1].find ("[", 1) + \
			1:result[1].find ("]", result[1].find ("[", 1))]
		self._lastUpdated = int (lineNumber)

	def PopDrugItem (self):
		"""
		Bridge.PopDrugItem (self)

		This will retrieve an item tacked onto the bridge from a drag
		operation.  On top of that, it also erases from the bridge in
		a traditional "pop" fashion.
		"""
		temp = self._drugItem
		self._drugItem = None
		return temp

	def SaveIPythonSession (self):
		"""
		Bridge.SaveIPythonSession (self)

		This method will create a File Saving dialog designed to save
		the contents of the IPython Shell's History to a log file. If
		the dialog is successful in getting a file name, it will save
		the entire contents of the history in clear text format, dis-
		regarding any formating that is actually seen in the IPython
		History.
		"""
		dialog = QFileDialog (None, QString ("Save IPython Session Log"),
			QString (str (QDir.home ().path ())),
			QString ("Text (*.log)"))
		dialog.setWindowIcon (QIcon (QString (ICON_DIR + "/file/python.png")))
		dialog.setViewMode (QFileDialog.Detail)
		dialog.setFileMode (QFileDialog.AnyFile)
		dialog.setAcceptMode (QFileDialog.AcceptSave)
		dialog.exec_ ()

		files = dialog.selectedFiles ()
		if not files.isEmpty ():
			fileName = str (files.takeFirst ())
			file = QFile ()
			file.setFileName (fileName)
			file.setPermissions (QFile.WriteOwner | QFile.ReadOwner | QFile.ReadUser \
				| QFile.WriteUser | QFile.ReadGroup | QFile.WriteGroup | QFile.ReadOther)
			if dialog.selectedFilter () == "Text (*.log)":
				file.open (QIODevice.WriteOnly | QIODevice.Text)
				file.writeData (str (self._ipyShell._commandHistory.toPlainText ()))
				file.close ()

	def Connect (self):
		"""
		Bridge.Connect (self)

		This method should be called whenever the user makes a
		successful connection.  This is done automatically through
		the Connection module.  It will enable the command line
		for ipython, make everything aware that it is connected, and
		set the menu actions to reflect connectivity.

		The only real way to connect is via the Connection module,
		so there is little need for this method to be called 
		anywhere else.
		"""
		self._ipyShell._commandLine.setReadOnly (False)
		self._ipyShell._commandStatus.toggleButton.setEnabled (True)
		self._connected = True
		self.pythonStatus ()
		self._actions.menuConnect ()

	def Disconnect (self):
		"""
		Bridge.Disconnect (self)

		This method should be called whenever the user disconnects
		from the basin session.  This can either happen manually by
		the user, or if the bridge's method Error deems it necessary
		to automatically kill the session.

		Generally speaking it looks for specific types of connectivity
		errors to decide if it should pull the plug on the session
		or if it will be ok.  Manually killing the basin_start.sh
		session, along with segfaults within the C-code will
		result in this type of disconnection.

		This method really only disables the ipython command line
		along with clearing the region/variable viewers, and changing
		menu's to reflect a disconnected status.
		"""
		self._ipyShell._commandLine.setReadOnly (True)
		self._ipyShell._initialized = False
		self._ipyShell._scriptPath = ""
		self._ipyShell._commandStatus.toggleButton.setEnabled (False)
		self._regionViewer.clear ()
		self._varViewer.clear ()
		self._connected = False
		self.pythonStatus ()
		self._actions.menuDisconnect ()

	def parallelOn (self):
		"""
		Bridge.parallelOn (self)
		
		This method simply makes the data viewers available as well
		as making the application aware that it is in parallel mode.

		This is not meant to change any behaviour, but rather reflect
		that state of the program.
		"""
		self._regionViewer.setEnabled (True)
		self._varViewer.setEnabled (True)
		self._parallel = True
		self.pythonStatus ()

	def parallelOff (self):
		"""
		Bridge.parallelOff (self)

		This method simply makes the data viewers disabled as well
		as informing the application as to its non-parallel status.

		This method will not affect any real behaviour, but rather
		reflects the state of the program.
		"""
		self._regionViewer.setEnabled (False)
		self._varViewer.setEnabled (False)
		self._parallel = False
		self.pythonStatus ()

	def pythonStatus (self):
		"""
		Bridge.pythonStatus (self)

		This method will update the IPython status, which is displayed
		in between the command history and command lines of the shell.

		The status string is dependant on the status of basin_client's
		connectivity as well as its parallel status.  Assuming it is
		both conencted and parallel, it will also inform the user of the
		number of nodes that it is currently working with in the session.
		"""
		if       self._parallel and     self._connected:
			self._ipyShell.showStatus ("Connected, Parallel Mode: On, " +
				str (len (self._nodes)) + " nodes")
		elif     self._parallel and not self._connected:
			self._ipyShell.showStatus ("Disconnected")
		elif not self._parallel and     self._connected:
			self._ipyShell.showStatus ("Connected, Parallel Mode: Off")
		elif not self._parallel and not self._connected:
			self._ipyShell.showStatus ("Disconnected")

	def AddToDock (self, item):
		"""
		Bridge.AddToDock (self, item)

		This method is generally called from the region or 
		free variable viwer.  It will send an item that it
		wants to send to the dock here first, since the viewers
		don't directly know which widget is currently docked
		(even though the user can obviously tell this).

		It will determine the proper widget in the dock to send
		the item to.  This does not imply that the dock can
		actually do anythung with the item, that is up to the
		dock widget to decide.
		"""
		if item != None:
			currentWidget = self._dock.GetCurrentWidget ()
			if currentWidget != None:
				currentWidget.DoubleClicked (item)

	def makeParallel (self):
		"""
		Bridge.makeParallel (self)

		This is just a method from the bridge to force the client into
		a parallel mode.  If it already is parallel, it does nothing.
		"""
		if self._parallel == False:
			self._ipyShell.toggleSilentAutopx ()

	def makeLocal (self):
		"""
		Bridge.makeParallel (self)

		This is just a method from the bridge to force the client into
		a non-parallel mode.  If it isn't parallel currently, it does nothing.
		"""
		if self._parallel == True:
			self._ipyShell.toggleSilentAutopx ()

	def timerEvent (self, ev):
		"""
		Bridge.timerEvent (self, ev)

		In the Bridge's initialization it sets a timer event, currently
		for every 3 seconds.  Every 3 seconds, this method is called.

		If the client is connected, parallel, and the user has nothing
		currently typed on the command line, it will go ahead and see
		if it needs to bother updating.

		In the bridge, there is a variable keeping track of the last
		updated line in the ipython shell.  It will check that against
		a variable that keeps track of the latest relevant update line
		in ipython.  This assumes that code in the ipython shell knows
		when it is and isn't relevant to updates.

		If it finds that its update line is less than the ipython's
		last relevant update line, then it will go ahead and perform
		an Update, as well as setting its newly updated lastUpdated
		variable.
		"""

		# Only attempt auto-updates if you are connected and parallel
		if self._parallel and self._connected and self._ipyShell.isClear () \
		and self._ipyShell._commandLine._singleLine: 
			# Retrieve the last update-worthy line number
			result = self._ipyShell.SilentCommand ("print \"\\n\" + BASIN_UPDATE")
			if result.find ("rror") != -1:
				self.Error ("BASIN_bridge.py", "Bridge", "timerEvent", result)

				# A connection error, most likely, so disconnect
				self.makeLocal ()
				self.Disconnect ()
				return
			val = result.split ("\n")[3]
			# And see if it is greater than your last updated line number
			if self._lastUpdated < int (val):
				# If they are, update your last updated number, and actually update
				self._lastUpdated = int (val)
				self.Update ()

	def Error (self, _file, _class, _function, _message = "", silent = False):
		"""
		Bridge.Error (self, _file, _class, _function, _message = "", silent)

		This error method should be called whenever ipython makes a boo-boo.
		It is up the the coder to determine if an error has occurred, I lazily
		do this by doing a quick search through the return string for error.

		All of the arguments are for the purpose of reporting where in this
		basin_client code that the error occurred in.  Of course the error
		may have nothing to do with the client code, however, it is possible
		that the error occurred from a mal-constructed ipython command being
		sent to the ipython shell, in which case it is important.  _file refers
		to the actual scrip file, _class to the python class it occurred in,
		_function for the method in the class it occurred.  _message should be
		a copy of the ipython error received.  Silent hasn't been implemented
		yet, but was intended to be a way of suppressing error messages to the
		user, while still logging them in ~/.basin/client/error.log
		"""
		self.CursorNorm ()

		# Retrieve Date & Time
		_date = QDate ().currentDate ().toString ("MM/dd/yyyy")
		_time = QTime ().currentTime ().toString ("H:mm:ss")

		# Log error
		path = QDir.home ()
		if not path.exists (".basin"):
			path.mkdir (".basin")
		path.cd (".basin")
		if not path.exists ("client"):
			path.mkdir ("client")
		path.cd ("client")
		logPath = path.path () + "/error.log"
		file = open (logPath, "a")
		file.write ("Date: " + str (_date) + "\n")
		file.write ("Time: " + str (_time) + "\n")
		file.write ("File: " + str (_file) + "\n")
		file.write ("Class: " + str (_class) + "\n")
		file.write ("Function: " + str (_function) + "\n\n")
		file.write ("***************************************************************************\n")
		file.write ("*******************************Error Message*******************************\n")
		file.write ("***************************************************************************\n")
		file.write (_message + "\n")
		file.write ("***************************************************************************\n\n\n\n\n")
		file.close ()

		if silent:
			return

		# Inform user of error
		dialog = QDialog (self._mainWindow)
		dialog.setWindowTitle ("Error!")
		layout = QGridLayout ()
		dialog.setLayout (layout)
		acceptButton = QPushButton ("Ok")
		QObject.connect (acceptButton, SIGNAL ("clicked ()"), dialog.accept)
		layout.addWidget (QLabel (QString ("IPython Error!")), 0, 0, 1, 2, Qt.AlignHCenter)
		layout.addWidget (QLabel (QString ("Date:")), 1, 0, 2, 1, Qt.AlignRight)
		layout.addWidget (QLabel (QString (str (_date))), 1, 1, 2, 1, Qt.AlignLeft)
		layout.addWidget (QLabel (QString ("Time:")), 3, 0, 2, 1, Qt.AlignRight)
		layout.addWidget (QLabel (QString (str (_time))), 3, 1, 2, 1, Qt.AlignLeft)
		layout.addWidget (QLabel (QString ("File:")), 5, 0, Qt.AlignRight)
		layout.addWidget (QLabel (QString (str (_file))), 5, 1, Qt.AlignLeft)
		layout.addWidget (QLabel (QString ("Class:")), 6, 0, Qt.AlignRight)
		layout.addWidget (QLabel (QString (str (_class))), 6, 1, Qt.AlignLeft)
		layout.addWidget (QLabel (QString ("Function:")), 7, 0, Qt.AlignRight)
		layout.addWidget (QLabel (QString (str (_function))), 7, 1, Qt.AlignLeft)
		layout.addWidget (QLabel (QString ("Log available in " + logPath)), \
			8, 0, 1, 2, Qt.AlignHCenter)
		layout.addWidget (acceptButton, 9, 0, 1, 2, Qt.AlignHCenter)
		dialog.exec_ ()
