"""
Overview
========
This module provides the dialog necessary for connecting to a Basin Remote session.  It
provides a mechanism for permanently storing profiles for a user to easily connect to
multiple sessions.
"""

from PyQt4.QtGui  import *
from PyQt4.QtCore import *

import re

import Constants

class ConnectionDialog (QDialog):
	"""
	"""

	def __init__ (self, bridge):
		"""
		"""

		QDialog.__init__ (self, bridge._mainWindow)
		self._bridge = bridge
		mainRect = self._bridge._mainWindow.geometry  ()
		width = mainRect.width ()
		height = mainRect.height ()
		x = mainRect.x () + int (width * 0.25)
		y = mainRect.y () + int (height * 0.40)
		width = int (width * 0.5)
		height = int (height * 0.3)
		self.setGeometry (x, y, width, height)
		layout = QVBoxLayout ()
		self.setLayout (layout)

		# A dictionary of profile names, to Profile objects
		self.profiles = {}

		# Set the window icon/title
		self.setWindowTitle (QString ("Connect..."))
		self.setWindowIcon (QIcon (Constants.ICON_DIR + "/file/connect.png"))

		profileSelectLayout = QHBoxLayout ()
		profileSelectLabel = QLabel (QString ("Profile:"))
		profileSelectLayout.addWidget (profileSelectLabel)
		self._profileSelect = QComboBox ()
		QObject.connect (self._profileSelect, \
			SIGNAL ("currentIndexChanged (const QString&)"), self.profileChanged)
		profileSelectLayout.addWidget (self._profileSelect)
		layout.addLayout (profileSelectLayout)

		self._profile = Constants.BTextField (layout, "Profile Name", "")
		self._profile.setEnabled (False)
		self._hostLine = Constants.BTextField (layout, "BASIN Host", "")
		self._portLine = Constants.BTextField (layout, "BASIN Port", "")
		self._userLine = Constants.BTextField (layout, "User Name", "")

		self.loadSettings ()

		connectBox = QGroupBox ("Connection Actions")
		connectBoxLayout = QHBoxLayout ()
		connectBox.setLayout (connectBoxLayout)
		profileBox = QGroupBox ("Profile Actions")
		profileBoxLayout = QHBoxLayout ()
		profileBox.setLayout (profileBoxLayout)

		# Everything concerning the "Connect" button
		connectButton = Constants.BQPushButton ("Connect", connectBoxLayout)
		connectButton.setToolTip ("Connect with current profile")
		QObject.connect (connectButton, SIGNAL("clicked ()"), self.accept)

		# Everything concerning the "Cancel" button
		cancelButton = Constants.BQPushButton ("Cancel", connectBoxLayout)
		cancelButton.setToolTip ("Exit dialog without connecting")
		QObject.connect (cancelButton, SIGNAL("clicked ()"), self.reject)


		# Everything concerning the "New" button
		newButton = Constants.BQPushButton ("New", profileBoxLayout)
		newButton.setToolTip ("Create new profile")
		QObject.connect (newButton, SIGNAL ("clicked ()"), self.newProfile)

		# Everything concerning the "Remove" button
		removeButton = Constants.BQPushButton ("Remove", profileBoxLayout)
		removeButton.setToolTip ("Remove current profile")
		QObject.connect (removeButton, SIGNAL ("clicked ()"), self.removeProfile)

		# Everything concerning the "Save" button
		saveButton = Constants.BQPushButton ("Save", profileBoxLayout)
		saveButton.setToolTip ("Permanently change settings")
		QObject.connect (saveButton, SIGNAL ("clicked ()"), self.saveProfile)

		# Accumulate all sub-layouts into the main layout
		layout.addWidget (connectBox)
		layout.addWidget (profileBox)

		# Actually start the dialog
		self.exec_ ()

	def profileChanged (self, string):
		"""
		"""

		self._profile.setEnabled (False)
		profile = str (string)
		self._profile.setText (profile)
		self._hostLine.setText (self.profiles [profile].getHost ())
		self._portLine.setText (self.profiles [profile].getPort ())
		self._userLine.setText (self.profiles [profile].getUser ())

	def createProfile (self, profile):
		"""
		"""

		# Creates new profile
		self.profiles[str (profile)] = Profile (str (profile))
		self._profileSelect.addItems (QStringList (str (profile)))

		# Jumps to new Profile
		index = self._profileSelect.findText (profile)
		self._profileSelect.setCurrentIndex (index)
		

	def newProfile (self):
		"""
		"""

		dialog = QDialog (self)
		dialog.setWindowTitle ("New Profile")
		layout = QVBoxLayout (dialog)
		profileName = Constants.BTextField (layout, "Profile Name", "")
		buttonLayout = Constants.BQHBoxLayout (layout)
		acceptButton = Constants.BQPushButton ("Ok", buttonLayout)
		QObject.connect (acceptButton, SIGNAL ("clicked ()"), dialog.accept)
		rejectButton = Constants.BQPushButton ("Cancel", buttonLayout)
		QObject.connect (rejectButton, SIGNAL ("clicked ()"), dialog.reject)
		dialog.exec_ ()

		if dialog.result () == 1:  # Accepted
			self.createProfile (profileName.text ())

	def saveProfile (self):
		"""
		Overview
		========
		This method should be called whenever you want to save updates to a profile.
		You may want to do this whether the profile is new and as yet unsaved, or if
		it is old and has just had changed applied to it.  Afterwards it will save the
		settings to file.
		"""

		self._profile.setEnabled (False)
		profile = str (self._profile.text ())

		if profile not in self.profiles:
			self.createProfile (profile)

		self.profiles[profile].setHost (str (self._hostLine.text ()))
		self.profiles[profile].setPort (str (self._portLine.text ()))
		self.profiles[profile].setUser (str (self._userLine.text ()))
		self.saveSettings ()

	def removeProfile (self):
		"""
		Overview
		========
		This method will remove the profile currently being viewed.  While this may
		seem permanent, this effect will not be permanent if the settings are not saved.
		"""

		# This will remove the currently viewed profile.
		# This isn't permanent until the user clicks save or connect.
		if self._profile.isEnabled ():
			return
		profile = str (self._profile.text ())
		del (self.profiles[profile])
		self._profileSelect.removeItem (self._profileSelect.currentIndex ())

	def loadSettings (self):
		"""
		"""

		file = QFile (str (QDir.home ().path ()) + "/.basin/client/connection.conf")
		if file.exists ():
			file.open (QIODevice.ReadOnly | QIODevice.Text)
			stream = QTextStream (file)
			default = None
			reDefault = re.compile ("\s*DEFAULT\s*:\s*[0-9]*\s*")
			reProfile = re.compile ("\s*PROFILE\s*:\s*.*\s*")
			reHost = re.compile ("\s*HOST\s*:\s*.*\s*:\s*.*\s*")
			rePort = re.compile ("\s*PORT\s*:\s*.*\s*:\s*.*\s*")
			reUser = re.compile ("\s*USER\s*:\s*.*\s*:\s*.*\s*")
			while not stream.atEnd ():
				line = str (stream.readLine ())

				if   reProfile.match (line):
					name = line[line.find (':') + 1:]
					if name not in self.profiles:
						self.profiles [name] = Profile (name)
				elif reHost.match (line):
					name = line[line.find (':') + 1:line.rfind (':')]
					host = line[line.rfind (':') + 1:]
					if name in self.profiles:
						self.profiles[name].setHost (host)
				elif rePort.match (line):
					name = line[line.find (':') + 1:line.rfind (':')]
					port = line[line.rfind (':') + 1:]
					if name in self.profiles:
						self.profiles[name].setPort (port)
				elif reUser.match (line):
					name = line[line.find (':') + 1:line.rfind (':')]
					user = line[line.rfind (':') + 1:]
					if name in self.profiles:
						self.profiles[name].setUser (user)
				elif reDefault.match (line):
					default = int (line[line.find (':') + 1:])
			file.close ()

			# Populate the profileSelector combobox with the profiles
			profileList = QStringList ()
			for profile in self.profiles.keys ():
				profileList.append (profile)
			self._profileSelect.addItems (profileList)

			# If a default profile was found and is valid, go to it
			if default and default < self._profileSelect.count ():
				self._profileSelect.setCurrentIndex (default)

	def saveSettings (self):
		"""
		Overview
		========
		This method will save all of the profiles created to file.  It will be saved under
		C{$HOME/.basin/client/connection.conf}
		"""

		path = QDir.home ()
		if not path.exists (".basin"):
			path.mkdir (".basin")
		path.cd (".basin")
		if not path.exists ("client"):
			path.mkdir ("client")
		path.cd ("client")

		file = QFile (str (path.path ()) + "/connection.conf")
		file.setPermissions (QFile.WriteOwner | QFile.ReadOwner | QFile.ReadUser | QFile.WriteUser)
		file.open (QIODevice.WriteOnly | QIODevice.Text)
		file.writeData ("Connection Preferences\n\n")

		file.writeData ("DEFAULT:" + str (self._profileSelect.currentIndex ()) + "\n\n")

		for profile in self.profiles.keys ():
			host = self.profiles[profile].getHost ()
			port = self.profiles[profile].getPort ()
			user = self.profiles[profile].getUser ()
			file.writeData ("PROFILE:" + profile + "\n")
			file.writeData ("HOST:" + profile + ":" + host + "\n")
			file.writeData ("PORT:" + profile + ":" + port + "\n")
			file.writeData ("USER:" + profile + ":" + user + "\n\n")
		file.close ()

	def accept (self):
		"""
		Overview
		========
		This method is called whenever "Connect" is pressed.  Incidently, it overrides the
		QDialog's accept method, however, calls the QDialog's accept within here.  It will
		save all of the current connection settings, and then send the host, port, and user
		to the ipython shell for connection.  During the course of all this the cursor
		will change to a busy status, and when it is done it will return.
		"""

		self._bridge.CursorBusy ()
		self.saveSettings ()
		self._bridge._host = self._hostLine.text ()
		self._bridge._port = self._portLine.text ()
		self._bridge._user = self._userLine.text ()
		self._bridge._ipyShell.Init (self._hostLine.text (), self._portLine.text (), \
			self._userLine.text ())
		QDialog.accept (self)
		self._bridge.CursorNorm ()


class Profile:
	"""
	Overview
	========
	This is just a small structure for keeping track of Connection Profiles. It only holds a name, host,
	port, and user. I use it by having a dictionary of these, where the profile name is also the dictionary key.
	"""

	def __init__ (self, name):
		"""
		@param name: The name of the profile.  This can't be changed later.
		@type name: string
		"""

		self._name = name
		self._host = ""
		self._port = ""
		self._user = ""

	def setHost (self, host):
		"""
		@param host: The name of the host to be set.
		@type host: string
		"""

		self._host = host

	def setPort (self, port):
		"""
		@param port: The port's number to be set.
		@type port: string
		"""

		self._port = port

	def setUser (self, user):
		"""
		@param user: The name of the user to be set.
		@type user: string
		"""

		self._user = user

	def getName (self):
		"""
		@return: string, the name of the profile.
		"""

		return self._name

	def getHost (self):
		"""
		@return: string, the name of the host.
		"""

		return self._host

	def getPort (self):
		"""
		@return: string, the port number.
		"""

		return self._port

	def getUser (self):
		"""
		@return: string, the name of the user.
		"""

		return self._user
