"""
Overview
========
This submodule provides specific implementations of L{DockWidget<client.Widget.DockWidget>}s for functions in Basin's Cosmology package.

There are currently three functions in the Cosmology package for basin that require a DockWidget.  They differ
greatly from each other, so unlike other package submodules, there is no abstract version of a Cosmology DockWidget.
"""

from PyQt4.QtGui import *
from PyQt4.QtCore import *

import re

import Constants
import DataGuesser
import Widget
import Holders

class Hubble (Widget.DockWidget):
	def __init__ (self, parent, bridge):
		self._type = "Cosmology"
		self._displayName = "Hubble Constant"
		self._displayIcon = QIcon (Constants.ICON_DIR + "/cosmology/hubble.png")
		self._description = "You may either supply a float or an attribute of floats. " + \
			"Your result will reflect the type of input you supplied."
		Widget.DockWidget.__init__ (self, parent, bridge)

	def _addCommands (self):
		box = QGroupBox ("Commands")
		layout = QHBoxLayout (box)

		self._action = Constants.BQPushButton ("Execute", layout)
		QObject.connect (self._action, SIGNAL ("clicked ()"), self.Execute)

		clearButton = Constants.BQPushButton ("Clear", layout)
		QObject.connect (clearButton, SIGNAL ("clicked ()"), self.clear)

		return box

	def _addDataSelection (self):
		self._in = AttributeHolder (self._bridge, self)
		self._in.setReadOnly (False)
		self._out = Holders.NameHolder ("hubble")

		self._dataType = QComboBox ()
		self._dataType.addItems (["Float", "Attribute"])
		QObject.connect (self._dataType, \
			SIGNAL ("currentIndexChanged (const QString&)"), self.changeType)

		outline = QGroupBox ("Data Selection")
		boxLayout = QGridLayout ()
		outline.setLayout (boxLayout)
		boxLayout.addWidget (QLabel (QString ("In/Out Type")), 0, 0, 1, 2, Qt.AlignRight)
		boxLayout.addWidget (self._dataType, 0, 2, 1, 2, Qt.AlignLeft)
		boxLayout.addWidget (QLabel (QString ("In:")), 1, 0, Qt.AlignRight)
		boxLayout.addWidget (self._in, 1, 1, Qt.AlignLeft)
		boxLayout.addWidget (QLabel (QString ("Out:")), 1, 2, Qt.AlignRight)
		boxLayout.addWidget (self._out, 1, 3, Qt.AlignLeft)

		return outline

	def clear (self):
		self._in.clear ()
		self._out.clear ()

	def changeType (self, text):
		self._in.clear ()
		current = self._dataType.currentIndex ()
		if   self._dataType.itemText (current) == "Attribute":
			self._in.setReadOnly (True)
		elif self._dataType.itemText (current) == "Float":
			self._in.setReadOnly (False)

	def Execute (self):
		if not ( self._bridge._connected == True and self._bridge._parallel == True):
			return False
		if self._in._shellAccess == None or self._out.text () == "":	
			return

		commands = []
		commands.append (str (self._out.text ()) + \
			" = COSMOLOGY.hubble (" + self._in._shellAccess + ")")
		self._bridge._ipyShell.BatchCommands (commands, True)

		self._in.clear ()
		self._out.clear ()

	def DoubleClicked (self, attr = None):
		if attr == None:
			attr = self._bridge._regionViewer.getSingleSelectedItem ()

		if attr.Type () == "ALIAS":
			attr = attr._item

		current = self._dataType.currentIndex ()
		if (attr.Type () == "ATTRIBUTE" or (attr.Type () == "VALUE" and \
		attr._type == "_basin.Attribute")) and self._dataType.itemText (current) == "Attribute":
			self._in._name = attr._name
			self._in._shellAccess = attr._shellAccess
			self._in.setText (attr._name)
		

class RedshiftToDistance (Widget.DockWidget):
	def __init__ (self, parent, bridge):
		self._type = "Cosmology"
		self._displayName = "Redshift To Distance"
		self._displayIcon = QIcon (Constants.ICON_DIR + "/cosmology/rdist.png")
		self._description = "some description here..."
		Widget.DockWidget.__init__ (self, parent, bridge)

	def _addCommands (self):
		box = QGroupBox ("Commands")
		layout = QHBoxLayout (box)

		self._action = Constants.BQPushButton ("Execute", layout)
		QObject.connect (self._action, SIGNAL ("clicked ()"), self.Execute)

		clearButton = Constants.BQPushButton ("Clear", layout)
		QObject.connect (clearButton, SIGNAL ("clicked ()"), self.clear)

		return box

	def _addOptions (self):
		box = QGroupBox ("Modes")
		layout = QHBoxLayout (box)

		self.pref = QCheckBox ("Execute in Place")
		QObject.connect (self.pref, SIGNAL ("stateChanged (int)"), self.setOutput)
		layout.addWidget (self.pref)

		return box

	def _addModes (self):
		box = QGroupBox ("Modes")
		layout = QVBoxLayout (box)

		distLabel = QLabel ("Distance Type:")
		self._distanceType = QComboBox ()
		self._distanceType.addItems (["Comoving", "Proper", "Angular Diameter", "Luminosity"])
		distLayout = Constants.BQHBoxLayout (layout)
		distLayout.addWidget (distLabel)
		distLayout.addWidget (self._distanceType)

		hLabel = QLabel ("h Result:")
		self._hType = QComboBox ()
		self._hType.addItems (["h^-1 Mpc", "h Mpc"])
		hLayout = Constants.BQHBoxLayout (layout)
		hLayout.addWidget (hLabel)
		hLayout.addWidget (self._hType)

		return box

	def _addDataSelection (self):
		self.list = Holders.ParentListHolder (self._bridge, ["redshift"], ["distance"])
		innerLayout = self.list.layout

		box = QGroupBox ("Data Selection")
		box.setLayout (innerLayout)
		return box

	def clear (self):
		self.list.clear ()

	def DefaultCheck (self):
		if self._bridge._inPlace:
			self.pref.setCheckState (Qt.Checked)
		else:
			self.pref.setCheckState (Qt.Unchecked)

	def guessDefaults (self):
		guessList = []
		guessList.append (DataGuesser.GenerateRegExpFromName ("Z"))
		DataGuesser.GuessAttributes (self, guessList)

	def setOutput (self, state):
		if   state == 0:
			self.list.setEnabled (False)
		elif state == 2:
			self.list.setEnabled (True)

	def Execute (self):
		if not ( self._bridge._connected == True and self._bridge._parallel == True):
			return False

		distFlag = self._distanceType.currentIndex ()
		hFlag = self._hType.currentIndex ()

		commands = []
		if   self.pref.checkState () == Qt.Unchecked:
			inAttribute = ""
			result = self.list.getInAccesses ()
			if result == False:
				return
			for attr in result:
				inAttribute += str (attr)
			outAttribute = ""
			result = self.list.getOutNames ()
			if result == False:
				return
			for attr in result:
				outAttribute += str (attr)

			commands.append (outAttribute + \
				" = COSMOLOGY.redshift_to_distance (" + str (inAttribute) + \
				", " + str (distFlag) + ", " + str (hFlag) + ")")

		elif self.pref.checkState () == Qt.Checked:
			inAttribute = ""
			result = self.list.getInNames ()
			if result == False:
				return
			for attr in result:
				inAttribute += ", \"" + str (attr) + "\""
			outAttribute = ""
			result = self.list.getOutNames ()
			if result == False:
				return
			for attr in result:
				outAttribute += ", \"" + str (attr) + "\""

			commands.append ("COSMOLOGY.redshift_to_distance (" + \
				str (self.list._shellAccess) + str (inAttribute) + \
				str (outAttribute) + ", " + str (distFlag) + ", " + str (hFlag) + ")")
		self._bridge._ipyShell.BatchCommands (commands, True)

		for holder in self.list._inList:
			holder.clear ()
		for holder in self.list._outList:
			holder.clear ()
		self.list.clear ()

	def DoubleClicked (self, attr = None):
		if attr == None:
			attr = self._bridge._regionViewer.getSingleSelectedItem ()

		if attr.Type () == "ALIAS":
			attr = attr._item

		if attr.Type () == "ATTRIBUTE" or (attr.Type () == "VALUE" \
		and attr._type == "_basin.Attribute"):
			for holder in self.list._inList:
				if holder.SetItem (attr, clobber = False) == True:
					return
		elif attr.Type () == "LIST":
			if self.list.text () == "None":
				self.list.setText (attr.text (0))
				self.list._shellAccess = attr._shellAccess

		
class RedshiftToGrowth (Widget.DockWidget):
	def __init__ (self, parent, bridge):
		self._type = "Cosmology"
		self._displayName = "Redshift To Growth"
		self._displayIcon = QIcon (Constants.ICON_DIR + "/cosmology/rgrowth.png")
		self._description = "some description here..."
		Widget.DockWidget.__init__ (self, parent, bridge)

	def _addOptions (self):
		box = QGroupBox ("Modes")
		layout = QHBoxLayout (box)

		self.pref = QCheckBox ("Execute in Place")
		QObject.connect (self.pref, SIGNAL ("stateChanged (int)"), self.setOutput)
		layout.addWidget (self.pref)

		return box


	def _addDataSelection (self):
		self.list = Holders.ParentListHolder (self._bridge, ["redshift"], ["growth"])
		innerLayout = self.list.layout

		box = QGroupBox ("Data Selection")
		box.setLayout (innerLayout)
		return box

	def _addCommands (self):
		box = QGroupBox ("Commands")
		layout = QHBoxLayout (box)

		self._action = Constants.BQPushButton ("Execute", layout)
		QObject.connect (self._action, SIGNAL ("clicked ()"), self.Execute)

		clearButton = Constants.BQPushButton ("Clear", layout)
		QObject.connect (clearButton, SIGNAL ("clicked ()"), self.clear)

		return box

	def clear (self):
		self.list.clear ()

	def DefaultCheck (self):
		if self._bridge._inPlace:
			self.pref.setCheckState (Qt.Checked)
		else:
			self.pref.setCheckState (Qt.Unchecked)

	def guessDefaults (self):
		guessList = []
		guessList.append (DataGuesser.GenerateRegExpFromName ("Z"))
		DataGuesser.GuessAttributes (self, guessList)

	def setOutput (self, state):
		if   state == 0:
			self.list.setEnabled (False)
		elif state == 2:
			self.list.setEnabled (True)

	def Execute (self):
		if not ( self._bridge._connected == True and self._bridge._parallel == True):
			return False

		commands = []
		if   self.pref.checkState () == Qt.Unchecked:
			inAttribute = ""
			result = self.list.getInAccesses ()
			if result == False:
				return
			for attr in result:
				inAttribute += str (attr)
			outAttribute = ""
			result = self.list.getOutNames ()
			if result == False:
				return
			for attr in result:
				outAttribute += str (attr)

			commands.append (outAttribute + " = COSMOLOGY.redshift_to_growth (" + 
				str (inAttribute) + ")")

		elif self.pref.checkState () == Qt.Checked:
			inAttribute = ""
			result = self.list.getInNames ()
			if result == False:
				return
			for attr in result:
				inAttribute += ", \"" + str (attr) + "\""
			outAttribute = ""
			result = self.list.getOutNames ()
			if result == False:
				return
			for attr in result:
				outAttribute += ", \"" + str (attr) + "\""

			commands.append ("COSMOLOGY.redshift_to_growth (" + \
				str (self.list._shellAccess) + str (inAttribute) + \
				str (outAttribute) + ")")
		self._bridge._ipyShell.BatchCommands (commands, True)

		for holder in self.list._inList:
			holder.clear ()
		for holder in self.list._outList:
			holder.clear ()
		self.list.clear ()

	def DoubleClicked (self, attr = None):
		if attr == None:
			attr = self._bridge._regionViewer.getSingleSelectedItem ()

		if attr.Type () == "ALIAS":
			attr = attr._item

		if attr.Type () == "ATTRIBUTE" or (attr.Type () == "VALUE" \
		and attr._type == "_basin.Attribute"):
			for holder in self.list._inList:
				if holder.SetItem (attr, clobber = False) == True:
					return
		elif attr.Type () == "LIST":
			if self.list.text () == "None":
				self.list.setText (attr.text (0))
				self.list._shellAccess = attr._shellAccess
		

class AttributeHolder (QLineEdit):
	def __init__ (self, bridge, parent):
		QLineEdit.__init__ (self, QString ("None"))
		self.parent = parent
		self._bridge = bridge
		self.setReadOnly (True)
		self.setAcceptDrops (True)
		self._name = None
		self._shellAccess = None
		QObject.connect (self, SIGNAL ("textEdited (const QString&)"), self.set)

	def set (self, text):
		self._name = str (text)
		self._shellAccess = str (text)

	def clear (self):
		self.setText ("None")
		self._name = None
		self._shellAccess = None

	def dropEvent (self, ev):
		data = self._bridge.PopDrugItem ()
		if not data:
			return
		if data.Type () == "ALIAS":
			data = data._item

		current = self.parent._dataType.currentIndex ()
		if   self.parent._dataType.itemText (current) == "Attribute":
			if   data.Type () == "ATTRIBUTE":
				self._name = data._name
				self._shellAccess = data._shellAccess
				self.setText (self._name)
				ev.setDropAction (Qt.LinkAction)
				ev.accept ()
			elif data.Type () == "VALUE" and data._type ==  "_basin.Attribute":
				self._name = data._name
				self._shellAccess = data._name
				self.setText (self._name)
				ev.setDropAction (Qt.LinkAction)
				ev.accept ()
			else:
				ev.ignore ()
		elif self.parent._dataType.itemText (current) == "Float":
			if data.Type () == "VALUE" and data._type ==  "float":
				self._name = data._name
				self._shellAccess = data._name
				self.setText (self._name)
				ev.setDropAction (Qt.LinkAction)
				ev.accept ()

     	def dragEnterEvent (self, ev):
		if ev.mimeData ().hasFormat ("text/plain"):
			ev.setDropAction (Qt.LinkAction)
			ev.accept ()
		else:
			ev.ignore ()

