"""
Overview
========
This submodule provides a common interface for Dock Widgets with default implementations.
Dock Widgets are those components in the upper right corner of the Basin Client's central widget.
They are the menus that are seemingly dynamically swapped out when their button is clicked.
"""

import PyQt4.QtCore
import PyQt4.QtGui

class DockWidget (PyQt4.QtGui.QWidget):
	"""
	This class provides an interface for DockWidgets in an effort to make their appearance consistent.
	
	Overview
	========
	This is the only class within the DockWidget submodule.  It provides an interface for all
	Dock Widgets with default implementations.  The private methods provided all create different
	sections of the layout.  They are intended to give a consistent feel to all Dock Widgets.  Currently 
	there are two categories of private methods for this class:

	Implemented
	-----------
	These methods will have a default behavior that will generally assume that the instance
	of DockWidget (or any inherited classes) has certain private variables instantiated.  If
	such private variables are not found, then a blank implementation is assumed.  These methods
	are not intended to be overridden as they provide a common feel among all Dock Widgets.  See
	specific private methods for what variables they require.

	Unimplemented
	-------------
	These methods are unimplemented for one of two reasons:
	  1. They were optional sections
	  2. They are meant to be very specific for the particular Dock Widget
	    
	That being said, you never have to implement them, in which case the space they would have taken
	will be filled by the other sections.
	
	Important
	=========
	These private methods are for internal use only.  You should only ever override their implementations,
	never explicity call them.  They are called in the constructor once for a reason.
	"""

	def __init__ (self, parent, bridge):
		"""
		This constructor for a DockWidget will create a consistent look and feel for the entire DockWidget.

		The constructor will create the following sections of a DockWidget based on
		  1. Pre-existing Private Variables
		  2. Reimplementation of _add* methods

		These are the sections of a DockWidget.  Please note that they are all optional, and if you don't
		want them to appear, overload the corresponding _add* methods to return None.
		  - Header
		  - Description
		  - Commands
		  - Modes
		  - Options
		  - Data Selection
		"""

		PyQt4.QtGui.QWidget.__init__ (self)
		self.setVisible (False)
		self._bridge = bridge
		bridge._dock.layout.addWidget (self)

		overallLayout = PyQt4.QtGui.QVBoxLayout ()
		self.setLayout (overallLayout)
		mainLayout = PyQt4.QtGui.QHBoxLayout ()

		# Add the Header
		tempLayout = self._addHeader ()
		if tempLayout:
			overallLayout.addLayout (tempLayout)
			overallLayout.setAlignment (tempLayout, PyQt4.QtCore.Qt.AlignCenter)

		overallLayout.addLayout (mainLayout)

		# Sets up the left-hand side
		leftLayout = PyQt4.QtGui.QVBoxLayout ()
		mainLayout.addLayout (leftLayout)
		mainLayout.setStretchFactor (leftLayout, 1)

		# Add the Description
		tempLayout = self._addDescription ()
		if tempLayout:
			leftLayout.addWidget (tempLayout)

		# Add the Options
		tempLayout = self._addOptions ()
		if tempLayout:
			leftLayout.addWidget (tempLayout)

		# Add the Commands
		tempLayout = self._addCommands ()
		if tempLayout:
			leftLayout.addWidget (tempLayout)

		# Sets up the right-hand side
		rightLayout = PyQt4.QtGui.QVBoxLayout ()
		mainLayout.addLayout (rightLayout)
		mainLayout.setStretchFactor (rightLayout, 1)

		# Add the Modes
		tempLayout = self._addModes ()
		if tempLayout:
			rightLayout.addWidget (tempLayout)

		# Add the Data Selection
		tempLayout = self._addDataSelection ()
		if tempLayout:
			rightLayout.addWidget (tempLayout)
		

	def _addHeader (self):
		"""
		This method provides the DockWidget a consistent look for a header.

		Overview
		========
		This method will provide the consistent header of a DockWidget.  If any of the three
		required variables do not appear, the constructor will ignore making the header.

		Pre-existing Private Variables
		==============================
		  - self._displayIcon = QIcon (icon of the function)
		  - self._displayName = string (name of the function)
		  - self._type = string (name of the package)

		@return: This function may return 1 of 2 things:
		  1. QLayout
		  2. None
		"""

		try:
			picLabel = PyQt4.QtGui.QLabel ()
			picLabel.setPixmap (self._displayIcon.pixmap (PyQt4.QtCore.QSize (22, 22)))
			header = PyQt4.QtGui.QLabel ("<big><b>" + self._displayName + "</b> (" + self._type + ")</big>")
			header.setTextFormat (PyQt4.QtCore.Qt.RichText)

			layout = PyQt4.QtGui.QHBoxLayout ()
			layout.addWidget (picLabel)
			layout.addWidget (header)
			layout.setAlignment (picLabel, PyQt4.QtCore.Qt.AlignVCenter | PyQt4.QtCore.Qt.AlignRight)
			layout.setAlignment (header, PyQt4.QtCore.Qt.AlignVCenter | PyQt4.QtCore.Qt.AlignLeft)
			return layout
		except (AttributeError):
			# if self._type, self._displayName, or self._displayIcon weren't found
			# then return nothing
			return

	def _addDescription (self):
		"""
		This method provides the DockWidget with a consistent look for the function description

		Overview
		========
		This method will provide a description for the DockWidget contingent on it finding the
		single required variable.  In the event that the variable is not found, the constructor
		calling the method will ignore making a description.

		Pre-existing Private Variable
		=============================
		 - self._description = string (verbose description of the function)

		@return: This function may return 1 of 2 things:
		  1. QWidget
		  2. None
		"""

		try:
			box = PyQt4.QtGui.QGroupBox ("Description")
			layout = PyQt4.QtGuiQHBoxLayout (box)
			description = PyQt4.QtGuiQLabel (self._description)
			description.setWordWrap (True)
			layout.addWidget (description)
			return box
		except (AttributeError):
			# if self._description wasn't found then return nothing
			return

	def _addCommands (self):
		"""
		The purpose of this method is to create a consistent look and feel for a DockWidget's command buttons.
		
		Overview
		========
		This method has no default implementation since the commands vary between DockWidgets.  While
		you are able to put anything into this reimplementation, it is recommended that it remains
		consistent with the rest of the Basin Client.  Generally speaking, commands are buttons that
		will execute the function belonging to the DockWidget.  Please note that this simply sets up
		the buttons and their actions, it does not perform the action of the commands.

		@return: This function may return 1 of 2 things:
		  1. QWidget
		  2. None
		"""

		pass

	def _addModes (self):
		"""
		The purpose of this method is to create a consistent look and feel for DockWidgets who need modes.

		Overview
		========
		This method has no default implementation since not all DockWidgets will require a mode.
		While you are able to put anything into the reimplementation of this method, it is recommended
		that they remain consistent with the rest of the Basin Client.  Generally speaking, modes
		are drawn up as a horizontal layout of QRadioButtons.  Again, while not required, the point of
		the modes is to swap out the I{Data Selection} for a different mode.

		@return: This function may return 1 of 2 things:
		  1. QWidget
		  2. None
		"""

		pass

	def _addOptions (self):
		"""
		The purpose of this method is to create a consistent look and feel for options contained in DockWidgets.

		Overview
		========
		This method has no default implementation since not all DockWidgets will require options.
		While you are able to put anything into the reimplementation of this method, it is recommended
		that tehy remain consistent wit hthe rest of the Basin Client.  Generally speaking, options
		are drawn up as a vertical layout of QCheckBoxes.  Options are considered to be minor changes to the
		overall purpose of the function.

		@return: This function may return 1 of 2 things:
		  1. QWidget
		  2. None
		"""

		pass

	def _addDataSelection (self):
		"""
		The purpose of the method is to create a place for any DockWidget to place the guts of the function.

		Overview
		========
		This method has no default implementation since it would be unique for any DockWidget.  There is no
		real default layout for any function's data selection.  Bearing that in mind, this is the place where
		the actual data that the function intends to utilize should go.  Utilize the L{Holders<Holders>}
		submodule for this.

		@return: This function may return 1 of 2 things:
		  1. QWidget
		  2. None
		"""

		pass
