from PyQt4.QtGui  import *
from PyQt4.QtCore import *

import re

import Constants
import Connection

class Preferences (QDialog):
	"""
	Overview
	========
	This is the overall preferences dialog.  It essentially has a left column of different types of
	preferences, and by clicking on these options you load (and by load, I mean uncover) the section
	of preferences you want to edit.  The right side is dedicated to the sections.

	Each section is otherwise self sufficient from finding the settings to saving them.  This class
	manages the sections, not the preferences themselves.
	"""

	def __init__ (self, bridge):
		"""
		Overview
		========
		The initialization method sets up how the preferences dialog will work, but nothing about
		the actual preferences.

		It sets up the eometry of the dialog box, then makes a left-hand list of the sections that
		can be selected.  After actually creating the secitons in their respective classes, it sets
		up the mechanism for switching between them.
		"""

		# This simply sets up the position and size of the dialog
		# Currently it is set to be centered to the main widget
		# and be half of its size.
		QDialog.__init__ (self)
		self._bridge = bridge
		mainRect = self._bridge._mainWindow.geometry  ()
		width = mainRect.width ()
		height = mainRect.height ()
		x = mainRect.x () + int (width * 0.25)
		y = mainRect.y () + int (height * 0.25)
		width = int (width * 0.5)
		height = int (height * 0.5)
		self.setGeometry (x, y, width, height)

		# Sets up the icon and title for the dialog
		self._current = None
		self.setWindowIcon (QIcon (Constants.ICON_DIR + "/file/preferences.png"))
		self.setWindowTitle ("Preferences")
		layout = QHBoxLayout (self)

		# The left hand is a list-widget which contains the selector for
		# Switching between sections, and the Done button to close the dialog
		leftLayout = Constants.BQVBoxLayout (layout)
		self.moduleSelect = QListWidget ()
		leftLayout.addWidget (self.moduleSelect)
		doneButton = Constants.BQPushButton ("Done", leftLayout)
		QObject.connect (doneButton, SIGNAL ("clicked ()"), self.accept)

		self.moduleDisplay = Constants.BQHBoxLayout (layout)
		layout.setStretchFactor (leftLayout, 1)
		layout.setStretchFactor (self.moduleDisplay, 3)
		QObject.connect (self.moduleSelect, SIGNAL ("itemClicked (QListWidgetItem *)"), self.changeDisplay)
		QObject.connect (self.moduleSelect, SIGNAL ("itemSelected (QListWidgetItem *)"), self.changeDisplay)
		self.moduleSelect.setResizeMode (QListView.Fixed)

		# These are the sections to be added...
		self.addGeneral ()
		self.addConnection ()
		self.addToolBar ()
		self.addAppearance ()

		self.exec_ ()


	def changeDisplay (self, item):
		"""
		Overview
		========
		This method will make the current displayed preference section invisible, and then make
		the passed `item` preferences widget visisble.  This is connected to slicking or selecting
		items from the left-hand list of preferences sections.
		"""

		if self._current:
			self._current.setVisible (False)
		self._current = item._display
		self._current.setVisible (True)

	def addConnection (self):
		"""
		Overview
		========
		This method simply encapsulates the adding of the Connection Preferences.  It is an
		organizational encapsulation, as nothing is really gained from taking this out of __init__.

		This sets up the ConnectionPreferences class, and then makes it inviisble until the user
		actually wants to use it.
		"""

		item = QListWidgetItem ("Connection")

		display = QWidget ()
		display.setVisible (False)
		display.setLayout (ConnectionPreferences (self._bridge))

		item._display = display
		self.moduleSelect.addItem (item)
		self.moduleDisplay.addWidget (display)

	def addToolBar (self):
		"""
		Overview
		========
		This method simply encapsulates the adding of the Tool Bar Preferences.  It is an organizational
		encapsulation, as nothing is really gained from taking this out of __init__.

		This sets up the ToolBarPreferences class, and then makes it inviisble until the user actually
		wants to use it.
		"""

		item = QListWidgetItem ("Tool Bar")

		display = QWidget ()
		display.setVisible (False)
		layout = QVBoxLayout (display)
		layout.addLayout (ToolBarPreferences (self._bridge))

		item._display = display
		self.moduleSelect.addItem (item)
		self.moduleDisplay.addWidget (display)

	def addAppearance (self):
		"""
		Overview
		========
		This method simply encapsulates the adding of the Appearance Preferences.  It is an organizational
		encapsulation, as nothing is really gained from taking this out of __init__.

		This sets up the AppearancePreferences class, and then makes it inviisble until the user actually
		wants to use it.
		"""

		item = QListWidgetItem ("Appearance")

		display = QWidget ()
		display.setVisible (False)
		layout = QVBoxLayout (display)
		layout.addLayout (AppearancePreferences (self._bridge))

		item._display = display
		self.moduleSelect.addItem (item)
		self.moduleDisplay.addWidget (display)

	def addGeneral (self):
		"""
		Overview
		========
		This method simply encapsulates the adding of the General Preferences.  It is an organizational
		encapsulation, as nothing is really gained from taking this out of __init__.

		This sets up the GeneralPreferences class, and then makes it inviisble until the user actually
		wants to use it.
		"""

		item = QListWidgetItem ("General")

		display = QWidget ()
		display.setVisible (False)
		layout = QVBoxLayout (display)
		layout.addLayout (GeneralPreferences (self._bridge))

		item._display = display
		self.moduleSelect.addItem (item)
		self.moduleDisplay.addWidget (display)
		item.setSelected (True)
		self.changeDisplay (item)


class GeneralPreferences (QVBoxLayout):
	"""
	Overview
	========
	These are essentially the miscellaneous options that weren't grouped well enough to warrant their
	own section.  They include such things as:
	 - Default save/open directory
	 - Reduction in cluster output
	 - Dock widget's auto-guessing feature
	 - Dock widget's auto-inplace or free flaoting feature

	These settings get saved in ~/.basin/client/general.conf
	"""

	def __init__ (self, bridge):
		"""
		GeneralPreferences.__init__ (self, bridge)

		The initialization of the General Preferences section
		of the Preferences dialog simply lays out a few simple
		options for the user.

		It sets up a place to enter a default path for open/saving
		to default to.

		It sets up three checkboxes for auto-guessing auto-inplace
		functions, and limiting output.
		"""
		QVBoxLayout.__init__ (self)
		self._bridge = bridge

		header = QLabel ("General Preferences")
		header.setAlignment (Qt.AlignHCenter)
		self.addWidget (header)

		dataPathLayout = Constants.BQHBoxLayout (self)
		dataPathLabel = QLabel ("Default Data Path:")
		dataPathLabel.setAlignment (Qt.AlignRight | Qt.AlignVCenter)
		dataPathLayout.addWidget (dataPathLabel)
		self._dataPath = QLineEdit ()
		dataPathLayout.addWidget (self._dataPath)

		self._inplaceCheck = QCheckBox ("Functions default output to parent,\n" + \
			"instead of free floating variables\n" + \
			"when possible.")
		self.addWidget (self._inplaceCheck)

		self._autoguessCheck = QCheckBox ("Functions will automatically try\n" + \
			"to guess input values.")
		self.addWidget (self._autoguessCheck)

		self._limitoutCheck = QCheckBox ("Limit output to 1 node, even\n" + \
			"when you are in parallel mode.")
		self.addWidget (self._limitoutCheck)

		saveButton = Constants.BQPushButton ("Save", self)
		QObject.connect (saveButton, SIGNAL ("clicked ()"), self.saveSettings)
		self.getValues ()

	def getValues (self):
		"""
		GeneralPreferences.getValues (self)

		This method is generally only to be called initially.  It
		retrieves the settings that it controls from the bridge.

		The widgets which are affected from these settings need to
		be aware of the settings from the bridge, so this informs
		the preferences menu of what is already set.

		If settings are then applied after this, they will both be
		saved to the bridge, where they will go into effect, but
		also to the conf file: ~/.basin/client/general.conf
		"""
		if self._bridge._inPlace:
			self._inplaceCheck.setCheckState (Qt.Checked)
		else:
			self._inplaceCheck.setCheckState (Qt.Unchecked)
		if self._bridge._autoGuess:
			self._autoguessCheck.setCheckState (Qt.Checked)
		else:
			self._autoguessCheck.setCheckState (Qt.Unchecked)
		if self._bridge._limitOut:
			self._limitoutCheck.setCheckState (Qt.Checked)
		else:
			self._limitoutCheck.setCheckState (Qt.Unchecked)
		if self._bridge._dataPath:
			self._dataPath.setText (self._bridge._dataPath)
		else:
			self._dataPath.setText ("None")

	def saveSettings (self):
		"""
		GeneralPreferences.saveSettings (self)

		This is a fairly straight forward method which takes
		the current settings, regarldess of what they are, were,
		or anything, saves them to ~/.basin/client/general.conf
		
		If the file didn't already exist, thats no problem and it will
		now.  This will also make the bridge re-load the values from the
		newly created file.

		In general, the format of the file ignores non newline whitespace.
		It follows the format::

		  SETTING:value

		Order is irrelevant!

		Everything obviously gets saved as a string, to this needs
		to be kept in mind when re-loading at the bridge.
		"""
		path = QDir.home ()
		if not path.exists (".basin"):
			path.mkdir (".basin")
		path.cd (".basin")
		if not path.exists ("client"):
			path.mkdir ("client")
		path.cd ("client")
		file = QFile (str (path.path ()) + "/general.conf" )
		file.setPermissions (QFile.WriteOwner | QFile.ReadOwner | QFile.ReadUser | QFile.WriteUser)
		file.open (QIODevice.WriteOnly | QIODevice.Text)

		file.writeData ("# General Preferences\n\n\n")

		file.writeData ("# True / False Section\n")
		if   self._inplaceCheck.checkState () == Qt.Unchecked:
			file.writeData ("INPLACE:False\n")
		elif self._inplaceCheck.checkState () == Qt.Checked:
			file.writeData ("INPLACE:True\n")

		if   self._autoguessCheck.checkState () == Qt.Unchecked:
			file.writeData ("AUTOGUESS:False\n")
		elif self._autoguessCheck.checkState () == Qt.Checked:
			file.writeData ("AUTOGUESS:True\n")

		if   self._limitoutCheck.checkState () == Qt.Unchecked:
			file.writeData ("LIMITOUT:False\n\n")
		elif self._limitoutCheck.checkState () == Qt.Checked:
			file.writeData ("LIMITOUT:True\n\n")

		file.writeData ("# A *NIX style absolute path\n")
		file.writeData ("DATAPATH:" + str (self._dataPath.text ()) + "\n\n")

		file.close ()

		self._bridge.loadSettings ()


class ToolBarPreferences (QVBoxLayout):
	"""
	Overview
	========
	This class is another section of settings that will dictate what actions are available to the user
	from the tool bar. At this point, they can select only entire packages to be displayed or removed from it.

	Also, currently there is just a single order that they are displayed with.  This means that File
	actions will always be at the beginning of the toolbar, Plotting next, etc, without being re-ordered.
	"""

	def __init__ (self, bridge):
		"""
		Overview
		========
		This initialization requires only the Bridge object to work properly.  It will essentially
		create 2 lists, one with the packages, and the other with the individial actions in each
		package.
		
		The list of actions is really just to inform the user what is exactly in each package.
		The list of packages, however, have checkboxes so that the user can choose which packages
		they would like on their toolbar.
		"""

		QVBoxLayout.__init__ (self)
		self._bridge = bridge
		self._toolBar = bridge._toolBar

		headerLayout = Constants.BQHBoxLayout (self)
		leftHeader = QLabel ("Categories")
		leftHeader.setAlignment (Qt.AlignHCenter)
		rightHeader = QLabel ("Contents")
		rightHeader.setAlignment (Qt.AlignHCenter)
		headerLayout.addWidget (leftHeader)
		headerLayout.addWidget (rightHeader)

		selectLayout = Constants.BQHBoxLayout (self)
		self.leftList = QListWidget ()
		self.rightList = QListWidget ()
		selectLayout.addWidget (self.leftList)
		selectLayout.addWidget (self.rightList)
		QObject.connect (self.leftList, SIGNAL ("itemActivated (QListWidgetItem *)"), self.updateCategory)
		QObject.connect (self.leftList, SIGNAL ("itemClicked (QListWidgetItem *)"), self.updateCategory)

		for category in self._toolBar.categories ():
			temp = QListWidgetItem (category[0])
			# For some reason this is the only way to enable 
			# checking in the first place...
			temp.setCheckState (Qt.Unchecked)
			if category[1]:
				temp.setCheckState (Qt.Checked)
			self.leftList.addItem (temp)

		saveButton = Constants.BQPushButton ("Save", self)
		QObject.connect (saveButton, SIGNAL ("clicked ()"), self.saveSettings)

	def updateCategory (self, item):
		"""
		Overview
		========
		This method is called whenever an item is clicked on the left-hand list, the package list.
		It will automatically update the right-hand action list to reflect the actions in the newly
		selected package.
		"""

		self.rightList.clear ()
		actions = self._toolBar.actions (item.text ())
		for action in actions:
			text = action.text ().replace ("&", "").replace ("...", "")
			if text != "":
				temp = QListWidgetItem (action.icon (), text)
				self.rightList.addItem (temp)

	def saveSettings (self):
		"""
		ToolBarPreferences.saveSettings (self)

		This method is called whenever settings are applied
		"""
		path = QDir.home ()
		if not path.exists (".basin"):
			path.mkdir (".basin")
		path.cd (".basin")
		if not path.exists ("client"):
			path.mkdir ("client")
		path.cd ("client")
		file = QFile (str (path.path ()) + "/menubar.conf" )
		file.setPermissions (QFile.WriteOwner | QFile.ReadOwner | QFile.ReadUser | QFile.WriteUser)
		file.open (QIODevice.WriteOnly | QIODevice.Text)
		file.writeData ("# Toolbar Preferences\n\n")
		file.writeData ("# True / False Section\n")

		for category in ["File", "Plotting", "General Functions", "Stellar Dynamics", \
		"Statistics", "Cosmology", "Linear Algebra", "Transformations", "Help"]:
			item = self.leftList.findItems (category, Qt.MatchExactly)[0]
			val = None
			if item.checkState () == Qt.Checked:
				val = "True"
			else:
				val = "False"

			if   item.text () == "File":
				file.writeData ("FILE:" + val + "\n")
			elif item.text () == "Plotting":
				file.writeData ("PLOT:" + val + "\n")
			elif item.text () == "General Functions":
				file.writeData ("GENERALFUNCTIONS:" + val + "\n")
			elif item.text () == "Stellar Dynamics":
				file.writeData ("STELLARDYNAMICS:" + val + "\n")
			elif item.text () == "Statistics":
				file.writeData ("STATISTICS:" + val + "\n")
			elif item.text () == "Cosmology":
				file.writeData ("COSMOLOGY:" + val + "\n")
			elif item.text () == "Linear Algebra":
				file.writeData ("LINEARALGEBRA:" + val + "\n")
			elif item.text () == "Transformations":
				file.writeData ("TRANSFORMATIONS:" + val + "\n")
			elif item.text () == "Help":
				file.writeData ("HELP:" + val + "\n")
		file.writeData ("\n")

		file.close ()
		self._toolBar.loadSettings ()


class AppearancePreferences (QGridLayout):
	def __init__ (self, bridge):
		QVBoxLayout.__init__ (self)
		self._bridge = bridge
		self._ipyShell = bridge._ipyShell._commandHistory

		# Create buffer values and populate them
		# These values will act as an intermediate for a user
		# to edit settings before they are committed to a both
		# the shell, and a conf file
		self._command = {}
		self._comment = {}
		self._result = {}
		self.PopulateBufferFromShell ()

		# The overall category selector
		selectLabel = QLabel ("Select:")
		selectLabel.setAlignment (Qt.AlignHCenter)
		self.categorySelect = QComboBox ()
		self.categorySelect.addItems (["Command", "Comment", "Result"])

		# The font family selector
		fontLabel = QLabel ("Font:")
		fontLabel.setAlignment (Qt.AlignHCenter)
		self.fontSelect = QFontComboBox ()

		# The font size spin-box
		sizeLabel = QLabel ("Size:")
		sizeLabel.setAlignment (Qt.AlignHCenter)
		self.sizeSelect = QSpinBox ()
		self.sizeSelect.setRange (1, 64)

		# The color selector and html code editor
		colorLabel = QLabel ("Color:")
		colorLabel.setAlignment (Qt.AlignHCenter)
		self.colorSelect = QComboBox ()
		self.colorSelect.addItem (QIcon (Constants.ICON_DIR + "/colors/white.png"), \
			"White", QVariant ("#ffffff"))
		self.colorSelect.addItem (QIcon (Constants.ICON_DIR + "/colors/red.png"), \
			"Red", QVariant ("#ff0000"))
		self.colorSelect.addItem (QIcon (Constants.ICON_DIR + "/colors/blue.png"), \
			"Blue", QVariant ("#000cff"))
		self.colorSelect.addItem (QIcon (Constants.ICON_DIR + "/colors/lightgreen.png"), \
			"Light Green", QVariant ("#19e83b"))
		self.colorSelect.addItem (QIcon (Constants.ICON_DIR + "/colors/darkgreen.png"), \
			"Dark Green", QVariant ("#05851a"))
		self.colorSelect.addItem (QIcon (Constants.ICON_DIR + "/colors/black.png"), \
			"Black", QVariant ("#000000"))
		self.colorSelect.addItem ("Custom", QVariant (""))
		colorcodeLabel = QLabel ("Color (HTML):")
		colorLabel.setAlignment (Qt.AlignHCenter)
		self.colorCode = QLineEdit ()

		# Area for selecting additional font decorations, bold, underline, etc
		decorationsBox = QGroupBox ("Decorations")
		decorationsLayout = QGridLayout (decorationsBox)
		self.bold = QCheckBox ("Bold")
		self.italics = QCheckBox ("Italics")
		self.underline = QCheckBox ("Underline")
		decorationsLayout.addWidget (self.bold, 0, 0)
		decorationsLayout.addWidget (self.italics, 1, 0)
		decorationsLayout.addWidget (self.underline, 2, 0)

		# Save button to save changes
		saveButton = QPushButton ("Save")
		saveButton.setAutoDefault (True)

		self.categorySelect.setCurrentIndex (1)

		QObject.connect (self.categorySelect, SIGNAL ("currentIndexChanged (const QString&)"), \
			self.updateCategory)
		QObject.connect (self.fontSelect, SIGNAL ("currentFontChanged (const QFont&)"), self.updateFont)
		QObject.connect (self.sizeSelect, SIGNAL ("valueChanged (int)"), self.updateSize)
		QObject.connect (self.colorSelect, SIGNAL ("currentIndexChanged (int)"), self.updateColor)
		QObject.connect (self.colorCode, SIGNAL ("editingFinished ()"), self.customColorEdit)
		QObject.connect (self.bold, SIGNAL ("stateChanged (int)"), self.updateBold)
		QObject.connect (self.italics, SIGNAL ("stateChanged (int)"), self.updateItalics)
		QObject.connect (self.underline, SIGNAL ("stateChanged (int)"), self.updateUnderline)
		QObject.connect (saveButton, SIGNAL ("clicked ()"), self.saveSettings)

		self.categorySelect.setCurrentIndex (0)

		# connect everything in one big lump
		self.addWidget (selectLabel, 1, 0, Qt.AlignRight)
		self.addWidget (self.categorySelect, 1, 1)
		self.addWidget (QLabel ("IPython Appearance"), 0, 0, 1, 2, Qt.AlignHCenter)
		self.addWidget (fontLabel, 2, 0, Qt.AlignRight)
		self.addWidget (self.fontSelect, 2, 1)
		self.addWidget (sizeLabel, 3, 0, Qt.AlignRight)
		self.addWidget (self.sizeSelect, 3, 1, Qt.AlignLeft)
		self.addWidget (colorLabel, 4, 0, Qt.AlignRight)
		self.addWidget (self.colorSelect, 4, 1)
		self.addWidget (colorcodeLabel, 5, 0, Qt.AlignRight)
		self.addWidget (self.colorCode, 5, 1, Qt.AlignLeft)
		self.addWidget (decorationsBox, 6, 0, 1, 2, Qt.AlignHCenter)
		self.addWidget (saveButton, 7, 0, 1, 2, Qt.AlignHCenter)

	def updateCategory (self, category):
		# Switch font name
		self.fontSelect.setCurrentFont (QFont (self.GetValFromBuffer (category, "Font")))

		# Switch font size
		self.sizeSelect.setValue (int (self.GetValFromBuffer (category, "Size")))

		# Switch font color
		colorIndex = self.colorSelect.findData (QVariant (self.GetValFromBuffer \
			(category, "Color")), Qt.UserRole, Qt.MatchExactly)
		if colorIndex == -1:
			# Set the color to custom and insert the html code
			self.colorSelect.setCurrentIndex (self.colorSelect.count () - 1)
			self.colorCode.setText (self.GetValFromBuffer (category, "Color"))
		else:
			# set the color to the known color, that should fill in the code automatically
			self.colorSelect.setCurrentIndex (colorIndex)

		# Switch font decorations
		self.bold.setCheckState (Qt.Unchecked)
		self.underline.setCheckState (Qt.Unchecked)
		self.italics.setCheckState (Qt.Unchecked)
		for dec in self.GetValFromBuffer (category, "Decorations"):
			if   dec == "b":
				self.bold.setCheckState (Qt.Checked)
			elif dec == "u":
				self.underline.setCheckState (Qt.Checked)
			elif dec == "i":
				self.italics.setCheckState (Qt.Checked)

	def updateFont (self, font):
		target = "Font"
		category = str (self.categorySelect.currentText ())
		fontName = str (font.family ())
		self.SendValToBuffer (category, target, fontName)

	def customColorEdit (self):
		color = self.colorCode.text ()
		category = str (self.categorySelect.currentText ())
		colorIndex = self.colorSelect.findData (QVariant (self.GetValFromBuffer \
			(category, "Color")), Qt.UserRole, Qt.MatchExactly)
		if colorIndex != -1:
			self.colorSelect.setCurrentIndex (colorIndex)
		else:
			self.colorSelect.setCurrentIndex (self.colorSelect.count () - 1)
			self.colorCode.setText (color)
		self.SendValToBuffer (category, "Color", color)

	def updateDecorations (self, state, val):
		target = "Decorations"
		category = str (self.categorySelect.currentText ())
		current = self.GetValFromBuffer (category, target)
		exists = current.count (val)
		final = current
		if   state == 0 and exists > 0:
			final = []
			for dec in current:
				if dec != val:
					final.append (dec)
		elif state == 2 and exists == 0:
			final.append (val)
			
		self.SendValToBuffer (category, target, final)

	def updateBold (self, state):
		self.updateDecorations (state, "b")

	def updateItalics (self, state):
		self.updateDecorations (state, "i")

	def updateUnderline (self, state):
		self.updateDecorations (state, "u")

	def updateSize (self, val):
		category = str (self.categorySelect.currentText ())
		target = "Size"
		self.SendValToBuffer (category, target, str (val))

	def updateColor (self, index):
		category = str (self.categorySelect.currentText ())
		target = "Color"
		if str (self.colorSelect.itemText (index)) == "Custom":
			self.colorCode.setText (self.GetValFromBuffer (category, target))
		else:
			variant = self.colorSelect.itemData (index)
			self.colorCode.setText (str (variant.toString ()))
			self.SendValToBuffer (category, target, str (variant.toString ()))

	def PopulateBufferFromShell (self):
		self.SendValToBuffer ("Command", "Font", self._ipyShell._commandFont)
		self.SendValToBuffer ("Command", "Size", self._ipyShell._commandSize)
		self.SendValToBuffer ("Command", "Color", self._ipyShell._commandColor)
		self.SendValToBuffer ("Command", "Decorations", self._ipyShell._commandDecorations)
		self.SendValToBuffer ("Comment", "Font", self._ipyShell._commentFont)
		self.SendValToBuffer ("Comment", "Size", self._ipyShell._commentSize)
		self.SendValToBuffer ("Comment", "Color", self._ipyShell._commentColor)
		self.SendValToBuffer ("Comment", "Decorations", self._ipyShell._commentDecorations)
		self.SendValToBuffer ("Result", "Font", self._ipyShell._resultFont)
		self.SendValToBuffer ("Result", "Size", self._ipyShell._resultSize)
		self.SendValToBuffer ("Result", "Color", self._ipyShell._resultColor)
		self.SendValToBuffer ("Result", "Decorations", self._ipyShell._resultDecorations)

	def CommitBufferToShell (self):
		self._ipyShell._commandFont = self.GetValFromBuffer ("Command", "Font")
		self._ipyShell._commandSize = self.GetValFromBuffer ("Command", "Size")
		self._ipyShell._commandColor = self.GetValFromBuffer ("Command", "Color")
		self._ipyShell._commandDecorations = self.GetValFromBuffer ("Command", "Decorations")
		self._ipyShell._commentFont = self.GetValFromBuffer ("Comment", "Font")
		self._ipyShell._commentSize = self.GetValFromBuffer ("Comment", "Size")
		self._ipyShell._commentColor = self.GetValFromBuffer ("Comment", "Color")
		self._ipyShell._commentDecorations = self.GetValFromBuffer ("Comment", "Decorations")
		self._ipyShell._resultFont = self.GetValFromBuffer ("Result", "Font")
		self._ipyShell._resultSize = self.GetValFromBuffer ("Result", "Size")
		self._ipyShell._resultColor = self.GetValFromBuffer ("Result", "Color")
		self._ipyShell._resultDecorations = self.GetValFromBuffer ("Result", "Decorations")

	def SendValToBuffer (self, target, field, value):
		if   target == "Command":
			target = self._command
		elif target == "Comment":
			target = self._comment
		elif target == "Result":
			target = self._result
		else:
			return

		if field != "Font" and field != "Size" and field != "Color" and field != "Decorations":
			return

		target[field] = value

	def GetValFromBuffer (self, target, field):
		if   target == "Command":
			target = self._command
		elif target == "Comment":
			target = self._comment
		elif target == "Result":
			target = self._result
		else:
			return

		if field != "Font" and field != "Size" and field != "Color" and field != "Decorations":
			return

		return target[field]

	def saveSettings (self):
		self.CommitBufferToShell ()

		path = QDir.home ()
		if not path.exists (".basin"):
			path.mkdir (".basin")
		path.cd (".basin")
		if not path.exists ("client"):
			path.mkdir ("client")
		path.cd ("client")

		file = QFile (str (path.path ()) + "/ipython_appearance.conf")
		file.setPermissions (QFile.WriteOwner | QFile.ReadOwner | QFile.ReadUser | QFile.WriteUser)
		file.open (QIODevice.WriteOnly | QIODevice.Text)
		file.writeData ("IPython Appearance Preferences\n\n")

		decs = ""
		for dec in self._ipyShell._commandDecorations:
			decs += dec + ","
		decs = decs[:-1]
		file.writeData ("COMMAND:SIZE:" + self._ipyShell._commandSize + "\n")
		file.writeData ("COMMAND:FONT:" + self._ipyShell._commandFont + "\n")
		file.writeData ("COMMAND:COLOR:" + self._ipyShell._commandColor + "\n")
		file.writeData ("COMMAND:DECORATIONS:" + decs + "\n\n")

		decs = ""
		for dec in self._ipyShell._resultDecorations:
			decs += dec + ","
		decs = decs[:-1]
		file.writeData ("RESULT:SIZE:" + self._ipyShell._resultSize + "\n")
		file.writeData ("RESULT:FONT:" + self._ipyShell._resultFont + "\n")
		file.writeData ("RESULT:COLOR:" + self._ipyShell._resultColor + "\n")
		file.writeData ("RESULT:DECORATIONS:" + decs + "\n\n")

		decs = ""
		for dec in self._ipyShell._commentDecorations:
			decs += dec + ","
		decs = decs[:-1]
		file.writeData ("COMMENT:SIZE:" + self._ipyShell._commentSize + "\n")
		file.writeData ("COMMENT:FONT:" + self._ipyShell._commentFont + "\n")
		file.writeData ("COMMENT:COLOR:" + self._ipyShell._commentColor + "\n")
		file.writeData ("COMMENT:DECORATIONS:" + decs + "\n\n")

		file.close ()
		self._ipyShell.loadSettings ()


class ConnectionPreferences (QVBoxLayout):
	def __init__ (self, bridge):
		QVBoxLayout.__init__ (self)
		self._bridge = bridge

		# A dictionary of profile names, to Profile objects
		self.profiles = {}

		profileSelectLayout = Constants.BQHBoxLayout (self)
		profileSelectLabel = QLabel (QString ("Profile:"))
		profileSelectLayout.addWidget (profileSelectLabel)
		self._profileSelect = QComboBox ()
		QObject.connect (self._profileSelect, \
			SIGNAL ("currentIndexChanged (const QString&)"), self.profileChanged)
		profileSelectLayout.addWidget (self._profileSelect)

		profileLayout = Constants.BQHBoxLayout (self)
		profileLabel = QLabel (QString ("Profile Name:"))
		profileLayout.addWidget (profileLabel)
		self._profile = QLineEdit ()
		self._profile.setEnabled (False)
		profileLayout.addWidget (self._profile)

		hostLayout = Constants.BQHBoxLayout (self)
		hostLabel = QLabel (QString ("BASIN Host:"))
		hostLayout.addWidget (hostLabel)
		self._hostLine = QLineEdit ()
		self._hostLine.setToolTip ("The hostname of your cluster")
		hostLayout.addWidget (self._hostLine)

		portLayout = Constants.BQHBoxLayout (self)
		portLabel = QLabel (QString ("BASIN Port:"))
		portLayout.addWidget (portLabel)
		self._portLine = QLineEdit ()
		self._portLine.setToolTip ("The port on your cluster")
		portLayout.addWidget (self._portLine)

		userLayout = Constants.BQHBoxLayout (self)
		userLabel = QLabel (QString ("User Name:"))
		userLayout.addWidget (userLabel)
		self._userLine = QLineEdit ()
		self._userLine.setToolTip ("")
		userLayout.addWidget (self._userLine)

		self.loadSettings ()

		# Start setting up the buttons
		buttonLayout = QGridLayout ()

		# Everything concerning the "New" button
		newButton = QPushButton ("New")
		buttonLayout.addWidget (newButton, 0, 0, 1, 2)
		newButton.setToolTip ("Create new profile")
		QObject.connect (newButton, SIGNAL ("clicked ()"), self.newProfile)

		# Everything concerning the "Remove" button
		removeButton = QPushButton ("Remove")
		buttonLayout.addWidget (removeButton, 0, 2, 1, 2)
		removeButton.setToolTip ("Remove current profile")
		QObject.connect (removeButton, SIGNAL ("clicked ()"), self.removeProfile)

		# Everything concerning the "Save" button
		saveButton = QPushButton ("Save")
		buttonLayout.addWidget (saveButton, 0, 4, 1, 2)
		saveButton.setToolTip ("Permanently change settings")
		QObject.connect (saveButton, SIGNAL ("clicked ()"), self.saveProfile)

		# Set "Connect" to be the default button
		#connectButton.setDefault (True)

		self.addLayout (buttonLayout)

	def profileChanged (self, string):
		self._profile.setEnabled (False)
		profile = str (string)
		self._profile.setText (profile)
		self._hostLine.setText (self.profiles [profile].getHost ())
		self._portLine.setText (self.profiles [profile].getPort ())
		self._userLine.setText (self.profiles [profile].getUser ())

	def newProfile (self):
		self._profile.setEnabled (True)
		self._profile.clear ()
		self._hostLine.clear ()
		self._portLine.clear ()
		self._userLine.clear ()

	def saveProfile (self):
		self._profile.setEnabled (False)
		profile = str (self._profile.text ())

		if profile not in self.profiles:
			self.profiles[profile] = Connection.Profile (profile)
			self._profileSelect.addItems (QStringList (profile))

		self.profiles[profile].setHost (str (self._hostLine.text ()))
		self.profiles[profile].setPort (str (self._portLine.text ()))
		self.profiles[profile].setUser (str (self._userLine.text ()))
		self.saveSettings ()

	def removeProfile (self):
		if self._profile.isEnabled ():
			return
		profile = str (self._profile.text ())
		del (self.profiles[profile])
		self._profileSelect.removeItem (self._profileSelect.currentIndex ())

	def loadSettings (self):
		file = QFile (str (QDir.home ().path ()) + "/.basin/client/connection.conf")
		if file.exists ():
			file.open (QIODevice.ReadOnly | QIODevice.Text)
			stream = QTextStream (file)
			reProfile = re.compile ("\s*PROFILE\s*:\s*.*\s*")
			reHost = re.compile ("\s*HOST\s*:\s*.*\s*:\s*.*\s*")
			rePort = re.compile ("\s*PORT\s*:\s*.*\s*:\s*.*\s*")
			reUser = re.compile ("\s*USER\s*:\s*.*\s*:\s*.*\s*")
			while not stream.atEnd ():
				line = str (stream.readLine ())

				if   reProfile.match (line):
					name = line[line.find (':') + 1:]
					if name not in self.profiles:
						self.profiles [name] = Connection.Profile (name)
				elif reHost.match (line):
					name = line[line.find (':') + 1:line.rfind (':')]
					host = line[line.rfind (':') + 1:]
					if name in self.profiles:
						self.profiles[name].setHost (host)
				elif rePort.match (line):
					name = line[line.find (':') + 1:line.rfind (':')]
					port = line[line.rfind (':') + 1:]
					if name in self.profiles:
						self.profiles[name].setPort (port)
				elif reUser.match (line):
					name = line[line.find (':') + 1:line.rfind (':')]
					user = line[line.rfind (':') + 1:]
					if name in self.profiles:
						self.profiles[name].setUser (user)
			file.close ()

			profileList = QStringList ()
			for profile in self.profiles.keys ():
				profileList.append (profile)
			self._profileSelect.addItems (profileList)


	def saveSettings (self):
		path = QDir.home ()
		if not path.exists (".basin"):
			path.mkdir (".basin")
		path.cd (".basin")
		if not path.exists ("client"):
			path.mkdir ("client")
		path.cd ("client")

		file = QFile (str (path.path ()) + "/connection.conf")
		file.setPermissions (QFile.WriteOwner | QFile.ReadOwner | QFile.ReadUser | QFile.WriteUser)
		file.open (QIODevice.WriteOnly | QIODevice.Text)
		file.writeData ("Connection Preferences\n\n")

		for profile in self.profiles.keys ():
			host = self.profiles[profile].getHost ()
			port = self.profiles[profile].getPort ()
			user = self.profiles[profile].getUser ()
			file.writeData ("PROFILE:" + profile + "\n")
			file.writeData ("HOST:" + profile + ":" + host + "\n")
			file.writeData ("PORT:" + profile + ":" + port + "\n")
			file.writeData ("USER:" + profile + ":" + user + "\n\n")
		file.close ()

