"""
Overview
========
This module adds the toolbar, a quicker and easier way for users to access the functions that they need.
Currently the toolbar is customizeable trhough the Preferences module.  It is allows the user to add
any combination of packages to the toolbar.  A package added to the toolbar adds all of the actions
from that package with it.  While there certainly could be a higher degree of flexibility, allowing
a user to manually edit whether they want every single menu option on their toolbar would be tedious.
"""

from PyQt4.QtGui  import *
from PyQt4.QtCore import *

import Constants
import re

class ToolBar (QToolBar):
	"""
	Overview
	========
	This class is the Basin Client's ToolBar.  It has the ability to add and remove entire packages
	of actions from it at a time.  It also stores these settings into a configuration file, so the
	same settings are persistent from use to use.  Currently, all the packages that exist can be
	placed into the toolbar.  They are as follows:
	 - File
	 - Plotting
	 - General Functions
	 - Stellar Dynamics
	 - Linear Algebra
	 - Statistics
	 - Cosmology
	 - Transformations
	 - Help
	"""

	def __init__ (self, parent, bridge):
		"""
		Overview
		========
		The initialization of ToolBar requires the parent
		object to be a QMainWindow, which it will embed itself in,
		and the bridge argument is the Bridge object which connects
		all of the interactive objects together.

		For each individual package of actions, this will take a list
		of references to each action from the bridge._actions object.
		This makes adding & deleting a group of actions to the toolbar
		as easy as iterating over a list.
		"""

		# Define basic toolbar information
		QToolBar.__init__ (self, parent)
		self.setIconSize (QSize (22, 22))
		self.setToolButtonStyle (Qt.ToolButtonIconOnly)
		self.setMovable (True)

		# Set the bridge, and have the bridge set ithe toolbar
		self._bridge = bridge
		bridge.setToolBar (self)

		# Define `File` button actions, and initialize them to being unset
		self._file = False
		self._fileActions = self._bridge._actions._file

		# Define `Plot` button actions, and initialize them to being unset
		self._plot = False
		self._plotActions = self._bridge._actions._plot

		# Define `General Functions` button actions, and initialize them to being unset
		self._generalfunctions = False
		self._generalfunctionsActions = self._bridge._actions._genf

		# Define `Stellar Dynamics` button actions, and initialize them to being unset
		self._stellardynamics = False
		self._stellardynamicsActions = self._bridge._actions._stel

		# Define `Linear Algebra` button actions, and initialize them to being unset
		self._linearalgebra = False
		self._linearalgebraActions = self._bridge._actions._linalg

		# Define `Statistics` button actions, and initialize them to being unset
		self._statistics = False
		self._statisticsActions = self._bridge._actions._stat

		# Define `Cosmology` button actions, and initialize them to being unset
		self._cosmology = False
		self._cosmologyActions = self._bridge._actions._cosm

		# Define `Transformations` button actions, and initialize them to being unset
		self._transformation = False
		self._transformationActions = self._bridge._actions._tran

		# Define `Help` button actions, and initialize them to being unset
		self._help = False
		self._helpActions = self._bridge._actions._help

		# Set the types you want, then make this the main window's toolbar
		parent.addToolBar (self)

	def categories (self):
		"""
		Overview
		========
		This method is used to get the names of all the categories, along with whether or not
		they are currently on your ToolBar.  This is useful for reporting to the Preferences
		modules which ones are currently selected.

		@return: This method will return a list of 2 element lists.  Each
			 inner 2 element list contains a string name, which is the
			 package of actions, and then a boolean which is True if the
			 package is being displayed in the ToolBar, False if not.
		"""
		return [["File", self._file], \
			["Plotting", self._plot], \
			["General Functions", self._generalfunctions], \
			["Cosmology", self._cosmology], \
			["Linear Algebra", self._linearalgebra], \
			["Stellar Dynamics", self._stellardynamics], \
			["Statistics", self._statistics], \
			["Transformations", self._transformation], \
			["Help", self._help]]

	def actions (self, category):
		"""
		Overview
		========
		This method will return the list of actions under a given
		package taken from the string argument, category.  This
		allows you to gather all of the actions in one fell swoop
		when you simply know the name of a package.

		@param category: The name of a package you want to retrieve the actions for.
		@type category: string
		"""
		if   category == "File":
			return self._fileActions
		elif category == "Plotting":
			return self._plotActions
		elif category == "General Functions":
			return self._generalfunctionsActions
		elif category == "Cosmology":
			return self._cosmologyActions
		elif category == "Linear Algebra":
			return self._linearalgebraActions
		elif category == "Stellar Dynamics":
			return self._stellardynamicsActions
		elif category == "Statistics":
			return self._statisticsActions
		elif category == "Transformations":
			return self._transformationActions
		elif category == "Help":
			return self._helpActions
		else:
			return []

	def clearSettings (self):
		"""
		Overview
		========
		THis method quite simply removes all actions from the toolbar, regardless of anything.  This is
		useful for getting hte toolbar back to a default state, especially to avoid adding a package
		twice to the ToolBar.
		"""

		self.remFileActions ()
		self.remPlotActions ()
		self.remGeneralFunctionsActions ()
		self.remStellarDynamicsActions ()
		self.remCosmologyActions ()
		self.remStatisticsActions ()
		self.remTransformationActions ()
		self.remLinearAlgebraActions ()
		self.remHelpActions ()

	def loadSettings (self):
		"""
		Overview
		========
		When this method is called, the configuration file ~/.basin/client/menubar.conf is loaded
		and parsed.  The configuration file should explicitly set or remove all packages from the
		toolbar.  This will make use of the L{_parseSetting<_parseSetting>} method which actually
		handles the adding or removal of the toolbar buttons.
		"""

		self.clearSettings ()
		file = QFile (str (QDir.home ().path ()) + "/.basin/client/menubar.conf")
		if file.exists ():
			file.open (QIODevice.ReadOnly | QIODevice.Text)
			stream = QTextStream (file)
			regexp = re.compile ("\s*.*\s*:\s*.*\s*")
			while not stream.atEnd ():
				line = str (stream.readLine ())
				if regexp.match (line):
					splitLine = line.replace (":", " ").split ()
					if len (splitLine) == 2:
						self._parseSetting (str (splitLine[0]), str (splitLine[1]))
			file.close ()
		else:
		# If the menubar.conf isn't even found, default the file actions to the toolbar
			self._parseSetting ("FILE", "True")

	def _parseSetting (self, key, value):
		"""
		Overview
		========
		This method is called from loadSettings so that it can parse the ~/.basin/client/menubar.conf
		configuration file.  The format is quite simple:

		C{PACKAGENAME:True|False}

		This method will end up calling the methods to add or remove the actions from the toolbar.  No harm
		will come from removing a package from the toolbar that isn't on it currently.

		@param key: An all caps representation of a package for Basin.  I.e. "FILE", "STELLARDYNAMICS", "HELP"...
		@type key: string

		@param value: Either "True" for a package that should be drawn to the toolbar, or anything else for false
		@type value: string
		"""

		if value != "True":
			return

		if   key == "FILE":
			self.addFileActions ()
		elif key == "PLOT":
			self.addPlotActions ()
		elif key == "GENERALFUNCTIONS":
			self.addGeneralFunctionsActions ()
		elif key == "STELLARDYNAMICS":
			self.addStellarDynamicsActions ()
		elif key == "STATISTICS":
			self.addStatisticsActions ()
		elif key == "COSMOLOGY":
			self.addCosmologyActions ()
		elif key == "LINEARALGEBRA":
			self.addLinearAlgebraActions ()
		elif key == "TRANSFORMATIONS":
			self.addTransformationActions ()
		elif key == "HELP":
			self.addHelpActions ()

	def addFileActions (self):
		"""
		Overview
		========
		Adds all of the I{File} actions from the toolbar.
		Actions may be added multiple times, though this would never be useful.
		"""

		for action in self._fileActions:
			self.addAction (action)
		self._file = True

	def remFileActions (self):
		"""
		Overview
		========
		Removes all of the I{File} actions from the toolbar.
		There is no harm in attempting to remove a package not currently on the toolbar.
		"""

		for action in self._fileActions:
			self.removeAction (action)
		self._file = False

	def addPlotActions (self):
		"""
		Overview
		========
		Adds all of the I{Plot} actions from the toolbar.
		Actions may be added multiple times, though this would never be useful.
		"""

		for action in self._plotActions:
			self.addAction (action)
		self._plot = True

	def remPlotActions (self):
		"""
		Overview
		========
		Removes all of the I{Plot} actions from the toolbar.
		There is no harm in attempting to remove a package not currently on the toolbar.
		"""

		for action in self._plotActions:
			self.removeAction (action)
		self._plot = False

	def addCosmologyActions (self):
		"""
		Overview
		========
		Adds all of the I{Cosmology} actions from the toolbar.
		Actions may be added multiple times, though this would never be useful.
		"""

		for action in self._cosmologyActions:
			self.addAction (action)
		self._cosmology = True

	def remCosmologyActions (self):
		"""
		Overview
		========
		Removes all of the I{Cosmology} actions from the toolbar.
		There is no harm in attempting to remove a package not currently on the toolbar.
		"""

		for action in self._cosmologyActions:
			self.removeAction (action)
		self._cosmology = False

	def addTransformationActions (self):
		"""
		Overview
		========
		Adds all of the I{Transformation} actions from the toolbar.
		Actions may be added multiple times, though this would never be useful.
		"""

		for action in self._transformationActions:
			self.addAction (action)
		self._transformation = True

	def remTransformationActions (self):
		"""
		Overview
		========
		Removes all of the I{Transformation} actions from the toolbar.
		There is no harm in attempting to remove a package not currently on the toolbar.
		"""

		for action in self._transformationActions:
			self.removeAction (action)
		self._transformation = False

	def addStellarDynamicsActions (self):
		"""
		Overview
		========
		Adds all of the I{Stellar Dynamics} actions from the toolbar.
		Actions may be added multiple times, though this would never be useful.
		"""

		for action in self._stellardynamicsActions:
			self.addAction (action)
		self._stellardynamics = True

	def remStellarDynamicsActions (self):
		"""
		Overview
		========
		Removes all of the I{Stellar Dynamics} actions from the toolbar.
		There is no harm in attempting to remove a package not currently on the toolbar.
		"""

		for action in self._stellardynamicsActions:
			self.removeAction (action)
		self._stellardynamics = False

	def addStatisticsActions (self):
		"""
		Overview
		========
		Adds all of the I{Statistics} actions from the toolbar.
		Actions may be added multiple times, though this would never be useful.
		"""

		for action in self._statisticsActions:
			self.addAction (action)
		self._statistics = True

	def remStatisticsActions (self):
		"""
		Overview
		========
		Removes all of the I{Statistics} actions from the toolbar.
		There is no harm in attempting to remove a package not currently on the toolbar.
		"""

		for action in self._statisticsActions:
			self.removeAction (action)
		self._statistics = False

	def addGeneralFunctionsActions (self):
		"""
		Overview
		========
		Adds all of the I{General Functions} actions from the toolbar.
		Actions may be added multiple times, though this would never be useful.
		"""

		for action in self._generalfunctionsActions:
			self.addAction (action)
		self._generalfunctions = True

	def remGeneralFunctionsActions (self):
		"""
		Overview
		========
		Removes all of the I{General Functions} actions from the toolbar.
		There is no harm in attempting to remove a package not currently on the toolbar.
		"""

		for action in self._generalfunctionsActions:
			self.removeAction (action)
		self._generalfunctions = False

	def addLinearAlgebraActions (self):
		"""
		Overview
		========
		Adds all of the I{Linear Algebra} actions from the toolbar.
		Actions may be added multiple times, though this would never be useful.
		"""

		for action in self._linearalgebraActions:
			self.addAction (action)
		self._linearalgebra = True

	def remLinearAlgebraActions (self):
		"""
		Overview
		========
		Removes all of the I{Linear Algebra} actions from the toolbar.
		There is no harm in attempting to remove a package not currently on the toolbar.
		"""

		for action in self._linearalgebraActions:
			self.removeAction (action)
		self._linearalgebra = False

	def addHelpActions (self):
		"""
		Overview
		========
		Adds all of the I{Help} actions from the toolbar.
		Actions may be added multiple times, though this would never be useful.
		"""

		for action in self._helpActions:
			self.addAction (action)
		self._help = True

	def remHelpActions (self):
		"""
		Overview
		========
		Removes all of the I{Help} actions from the toolbar.
		There is no harm in attempting to remove a package not currently on the toolbar.
		"""

		for action in self._helpActions:
			self.removeAction (action)
		self._help = False
