#!/usr/bin/env python

import sys
import PyQt4.QtGui
import PyQt4.QtCore

import Constants
import IPythonShell

import Actions
import Bridge
import Dock
import MenuBar
import StatusBar
import ToolBar
import Viewers

class BasinBody (PyQt4.QtGui.QWidget):
	"""
	Overview
	========
	This class in a convenience class which inherits from I{QWidget}.  It bundles the layout and instantiation
	of the main components of the central body widget during construction.  In every other aspect it behaves
	as a normal QWidget.
	"""

	def __init__ (self, parent, bridge):
		"""
		Overview
		========
		This constructor creates the central widget for the I{Basin Client} User Interface.  It
		is everything between the menu bar and status bar.  It will set up the following components:
		  - L{BasinViewer<client.Viewers.BasinViewer>}
		  - L{VariableViewer<client.Viewers.VariableViewer>}
		  - L{Dock<client.Dock.Dock>}
		  - L{IPythonView<client.IPythonShell.IPythonView>}

		@param parent: This is the main window which this body will be inserted into as the central component.
		@type parent: BasinWindow

		@param bridge: This is the instance of the bridge, the abstract component linking the
		interactive components
		@type bridge: Bridge
		"""

		# setup the widget object and its layout
		PyQt4.QtGui.QWidget.__init__ (self, parent)
		self.layout = PyQt4.QtGui.QHBoxLayout (self)
		splitter = PyQt4.QtGui.QSplitter ()
		self.layout.addWidget (splitter)

		# make a sub-layout of the left-hand side
		# which consists of the region & variable viewers
		leftSplitter = PyQt4.QtGui.QSplitter ()
		leftSplitter.setOrientation (PyQt4.QtCore.Qt.Vertical)
		self._regionViewer = Viewers.BasinViewer (leftSplitter, self, bridge)
		self._varViewer = Viewers.VariableViewer (leftSplitter, bridge)
		splitter.addWidget (leftSplitter)

		# right-hand layout, the dock and ipython shell
		rightSplitter = PyQt4.QtGui.QSplitter ()
		rightSplitter.setOrientation (PyQt4.QtCore.Qt.Vertical)
		self._dock = Dock.Dock (rightSplitter, bridge)
		self._ipyShell = IPythonShell.IPythonView (rightSplitter, bridge)
		splitter.addWidget (rightSplitter)

		# makes it so this created layout goes on the widget,
		# which becomes the main centeral widget
		self.setLayout (self.layout)
		parent.setCentralWidget (self)


class BasinWindow (PyQt4.QtGui.QMainWindow):
	"""
	Overview
	========
	This is simply a convenience class which inherits from I{QMainWindow}.  In every other way it behaves
	like a normal QMainWindow should.  The purpose of it is to wrap the necessary components
	of the I{Basin Client} User Interace into the window.  The window must be comprised of these elements,
	so it is fitting for them to be bundled into this class.
	"""

	def __init__ (self, app):
		"""
		Overview
		========
		This constructor will create the entire main window for the I{Basin Client}.
		It takes care of creating all the layouts and widgets for the entire application.
		It will not take care of showing and executing the main window, so the calling
		function should handle this.

		First, the parent class' constructor is called.  Then constructor sets the Basin title and icon.
		After this it instantiates the following components (order is important here):
		  1. L{Bridge<client.Bridge.Bridge>}
		  2. L{StatusBar<client.StatusBar.StatusBar>}
		  3. L{BasinBody<client.BasinBody>}
		  4. L{Actions<client.Actions.Actions>}
		  5. L{ToolBar<client.ToolBar.ToolBar>}
		  6. L{MenuBar<client.MenuBar.MenuBar>}

		@param app: This is the application for which you want this main window applied to.
		@type app: PyQt4.QtGui.QApplication
		"""

		# Normal setup of a QMainWindow with an icon and title
		PyQt4.QtGui.QMainWindow.__init__ (self)
		self.setWindowIcon (PyQt4.QtGui.QIcon (Constants.ICON_DIR + "/basin.png"))
		self.setWindowTitle ("BASIN")

		# The main parts of the main window,
		# I'll go into detail on their individual objects...
		self._bridge = Bridge.Bridge (app, self)
		self._status = StatusBar.StatusBar (self, self._bridge)
		self._body = BasinBody (self, self._bridge)
		self._actions = Actions.Actions (self._bridge)

		# Note that ToolBar and MenuBar depends on BasinActions,
		# They must be initialized AFTER the bridge.
		self._tool = ToolBar.ToolBar (self, self._bridge)
		self._menu = MenuBar.MenuBar (self, self._bridge)


def main (args):
	"""
	Overview
	========
	This simple main function creates the entire I{Basin Client} application.  When creating the
	instance of a QApplication, it currently does not make use of the command line arguments which
	are passed to this function.  Currently they have would have no context.  The function creates
	a working window for the application, shows the window and executes the application.

	@param args: This is a list of the command line arguments specified when running this script.
	@type args: list
	"""
	# The actual application object
	app = PyQt4.QtGui.QApplication ([], 1)

	# Set up the main window within the application
	win = BasinWindow (app)

	# win.show () for non-maximized appearance
	win.showMaximized ()

	# Starts main event loop
	app.exec_()


if __name__ == "__main__":
	main (sys.argv)
